#!/bin/bash
set -e

export APP_HOME=/app
export HOME=/app_home
export NUM_WORKERS=${NUM_WORKERS:-$((2 * $(nproc --all)))}

if [ "$1" == "version" ]; then
    exec python3 $APP_HOME/manage.py appversion
fi

RETRIES=10
until postgres_healthcheck; do
if [[ ${RETRIES} -eq 0 ]]; then
    echo "\nFailed to connect to database."
    exit 1
fi
echo -en "\nWaiting for database readiness ... "
sleep 1
RETRIES=$((RETRIES-1))
done

cd $APP_HOME
CONTAINER_MIGRATED="CONTAINER_MIGRATED_FLAG"
if [[ ! -e "$HOME/$CONTAINER_MIGRATED" && "$DJANGO_MIGRATE" == "yes"  ]]; then
    python3 $APP_HOME/manage.py migrate --noinput
    touch $HOME/$CONTAINER_MIGRATED
fi

CONTAINER_FIXTURE_APPLIED="CONTAINER_FIXTURE_APPLIED_FLAG"
if [ ! -e "$HOME/$CONTAINER_FIXTURE_APPLIED" -a -n "$DJANGO_LOAD_FIXTURE" -a "$DJANGO_LOAD_FIXTURE" != "no" ]; then
    python3 $APP_HOME/manage.py loaddata "$DJANGO_LOAD_FIXTURE"
    touch $HOME/$CONTAINER_FIXTURE_APPLIED
fi

ADMIN_CREATED="ADMIN_CREATED_FLAG"
if [ ! -e "$HOME/$ADMIN_CREATED" -a -n "$DJANGO_CREATE_ADMIN_PASSWORD" ]; then
    echo "Creating superuser '${DJANGO_CREATE_ADMIN_USERNAME:-admin}' ..."
    echo "from django.contrib.auth import get_user_model; User = get_user_model(); User.objects.filter(username='${DJANGO_CREATE_ADMIN_USERNAME:-admin}').exists() or User.objects.create_superuser('${DJANGO_CREATE_ADMIN_USERNAME:-admin}', 'noreply@localhost', '$DJANGO_CREATE_ADMIN_PASSWORD')" | python3 $APP_HOME/manage.py shell
    if [ -n "$SELECTED_CONFERENCE_ID" ]; then
        echo "Adding user as member of conference '$SELECTED_CONFERENCE_ID' ..."
        echo "from core.models import ConferenceMember, PlatformUser; u = PlatformUser.objects.get(username='${DJANGO_CREATE_ADMIN_USERNAME:-admin}'); ConferenceMember.objects.get_or_create(user=u, conference_id='${SELECTED_CONFERENCE_ID}', is_staff=True)" | python3 $APP_HOME/manage.py shell
    fi
    unset DJANGO_CREATE_ADMIN_PASSWORD
    touch $HOME/$ADMIN_CREATED
fi

if [ "$1" == "webworker" -o "$STARTMODE" == "webworker" ]; then
    # start gunicorn (might be called indirectly via supervisor when started with 'all')
    gunicorn_params=()
    [[ "${BIND_HTTP}" == "1" ]] && gunicorn_params+=(--bind=0.0.0.0:8000)
    if [[ "${DEV}" == "1" ]]; then
        #  Django deprecation warnings are sent to this logger, therefore we should watch out for it
        export PYTHONWARNINGS=always
        shopt -s globstar
        html_files=(**/*.html)
        gunicorn_params+=(--reload ${html_files[@]/#/--reload-extra-file })
    fi
    [[ -v GUNICORN_TIMEOUT ]] && gunicorn_params+=(--timeout "${GUNICORN_TIMEOUT}")
    exec gunicorn hub.wsgi \
        --name hub \
        --workers $NUM_WORKERS \
        --worker-tmp-dir /dev/shm \
        --max-requests 1200 \
        --max-requests-jitter 50 \
        --log-level=info \
        --access-logfile=/dev/null \
        --error-logfile=- \
        --bind=unix:/run/hub/app.sock \
        "${gunicorn_params[@]}"
fi

if [ "$1" == "shell" -o "$1" == "createsuperuser" -o "$1" == "test" ]; then
    exec python3 $APP_HOME/manage.py $1
fi

if [ "$1" == "housekeeping" ]; then
    interval="${HOUSEKEEPING_SLEEP_SECONDS:-300}"
    if [ "$interval" -gt 0 ]; then
        python3 $APP_HOME/manage.py housekeeping --forever --forever-delay="$interval"
    else
        python3 $APP_HOME/manage.py housekeeping
    fi
fi

cat <<EOD
Specify argument: webworker|shell|createsuperuser
EOD
exit 1
