import json
import uuid
from datetime import datetime
from zoneinfo import ZoneInfo

from django.contrib.gis.geos import Point
from django.test import TestCase, override_settings
from django.urls import reverse

from core.models import Assembly, Conference, ConferenceExportCache, PlatformUser

TEST_CONF_ID = uuid.uuid4()


@override_settings(SELECTED_CONFERENCE_ID=TEST_CONF_ID)
class MapTest(TestCase):
    def setUp(self):
        tz = ZoneInfo('CET')
        self.conf = Conference(
            id=TEST_CONF_ID,
            slug='conf',
            name='TestConf',
            is_public=True,
            start=datetime(2020, 12, 27, 1, 23, 45, tzinfo=tz),
            end=datetime(2020, 12, 30, 23, 45, 00, tzinfo=tz),
        )
        self.conf.save()
        self.conf.tracks.create(slug='community', name='Community').save()
        self.conf.tracks.create(slug='security', name='Security').save()
        self.assembly = Assembly(name='TestAssembly', slug='asmbly', conference=self.conf)
        self.assembly.save()

        self.user = PlatformUser(username='bernd', is_active=True)
        self.user.save()

    def test_map_poi(self):
        url = reverse('api:map-assemblies-poi')
        resp = self.client.get(url)
        self.assertEqual(200, resp.status_code)
        self.assertEqual(resp['Content-Type'], 'application/geo+json')

        # decode JSON
        content = json.loads(resp.content.decode('utf-8'))

        # check that there are no POIs yet
        self.assertEqual('FeatureCollection', content.get('type'))
        self.assertIsInstance(content.get('features'), list)
        self.assertEqual(0, len(content['features']))

        # set POI on assembly and accept it
        self.assembly.location_point = Point(8.9, 52.9)
        self.assembly.state = Assembly.State.ACCEPTED
        self.assembly.save()
        ConferenceExportCache.objects.all().delete()

        # check again, there should be no POI exported as the assembly is not placed yet
        resp = self.client.get(url)
        self.assertEqual(200, resp.status_code)
        content = json.loads(resp.content.decode('utf-8'))
        self.assertEqual(0, len(content['features']))

        # change state to 'placed', now there should be a POI
        self.assembly.state = Assembly.State.PLACED
        self.assembly.save()
        ConferenceExportCache.objects.all().delete()
        resp = self.client.get(url)
        self.assertEqual(200, resp.status_code)
        content = json.loads(resp.content.decode('utf-8'))
        self.assertEqual(1, len(content['features']))

        # check the POI
        feature = content['features'][0]
        self.assertEqual('Feature', feature.get('type'))
        self.assertEqual('Point', feature.get('geometry', {}).get('type'))
        self.assertListEqual([8.9, 52.9], feature.get('geometry', {}).get('coordinates'))
        self.assertEqual(self.assembly.name, feature.get('properties', {}).get('name'))
        self.assertEqual(self.assembly.is_cluster, feature.get('properties', {}).get('cluster'))

        # change assembly to 'hidden', POI should still be there
        self.assembly.state = Assembly.State.HIDDEN
        self.assembly.save()
        ConferenceExportCache.objects.all().delete()
        resp = self.client.get(url)
        self.assertEqual(200, resp.status_code)
        content = json.loads(resp.content.decode('utf-8'))
        self.assertEqual(1, len(content['features']))
