from datetime import UTC, datetime, timedelta
from unittest.mock import patch

from django.urls import reverse
from django.utils import timezone

from core.models import (
    ActivityLogEntry,
    Assembly,
    AssemblyMember,
    ConferenceMember,
    Invitation,
    PlatformUser,
)
from core.tests.mock import mocktrans

from backoffice.tests.base import BackOfficeTestCase


class InvitationHabitatTestCase(BackOfficeTestCase):
    def setUp(self):
        super().setUp()

        self.assembly = Assembly.objects.create(name='Assembly 1', slug='assembly_1', conference=self.conf)
        self.assembly_manager = PlatformUser.objects.create(username='assembly_manager')
        self.assembly_manager_cm = ConferenceMember.objects.create(conference=self.conf, user=self.assembly_manager)
        self.assembly_manager_am = AssemblyMember.objects.create(
            assembly=self.assembly,
            member=self.assembly_manager,
            is_representative=True,
            can_manage_assembly=True,
        )
        self.habitat = Assembly.objects.create(name='Habitat 1', slug='habitat_1', conference=self.conf, hierarchy=Assembly.Hierarchy.CLUSTER)
        self.habitat_manager = PlatformUser.objects.create(username='habitat_manager')
        self.habitat_manager_cm = ConferenceMember.objects.create(conference=self.conf, user=self.habitat_manager)
        self.habitat_manager_am = AssemblyMember.objects.create(
            assembly=self.habitat,
            member=self.habitat_manager,
            is_representative=True,
            can_manage_assembly=True,
        )


class InvitationHabitatSendViewTestCase(InvitationHabitatTestCase):
    def test_missing_requester_id(self):
        self.client.force_login(self.habitat_manager)
        response = self.client.post(
            reverse('backoffice:invitation-send', kwargs={'type': 'habitat'}),
        )
        self.assertContains(response, 'Requester ID is required', status_code=400)

    def test_send_habitat_assembly_invitation_accept(self):
        self.client.force_login(self.habitat_manager)
        # Send invitation from habitat to assembly
        response = self.client.post(
            reverse(
                'backoffice:invitation-send-with-pk',
                kwargs={
                    'type': 'habitat',
                    'requester_id': self.habitat.id,
                },
            ),
            data={
                'requester_id': self.habitat.id,
                'requested_id': self.assembly.id,
            },
        )
        self.assertEqual(Invitation.objects.count(), 1)
        invitation = Invitation.objects.first()
        self.assertEqual(invitation.sender, self.habitat_manager)
        self.assertEqual(invitation.type, Invitation.InvitationType.HABITAT)
        self.assertRedirects(response, reverse('backoffice:invitation', kwargs={'pk': invitation.id}))

        # Accept invitation
        self.client.force_login(self.assembly_manager)
        self.assertIsNone(self.assembly.parent)
        activity_log_count = ActivityLogEntry.objects.count()
        response = self.client.post(
            reverse('backoffice:invitation', kwargs={'pk': invitation.id}),
            data={
                'action': 'accept',
            },
        )
        self.assertRedirects(response, reverse('backoffice:invitation', kwargs={'pk': invitation.id}))
        self.assembly.refresh_from_db()
        invitation.refresh_from_db()

        self.assertEqual(invitation.decision_by, self.assembly_manager)
        self.assertEqual(invitation.state, Invitation.RequestsState.ACCEPTED)

        self.assertEqual(self.assembly.parent, self.habitat)
        self.assertEqual(ActivityLogEntry.objects.count(), activity_log_count + 2)

    def test_send_assembly_habitat_invitation_accept(self):
        self.client.force_login(self.assembly_manager)
        # Send invitation from habitat to assembly
        response = self.client.post(
            reverse(
                'backoffice:invitation-send-with-pk',
                kwargs={
                    'type': 'habitat',
                    'requester_id': self.assembly.id,
                },
            ),
            data={
                'requester_id': self.assembly.id,
                'requested_id': self.habitat.id,
            },
        )
        self.assertEqual(Invitation.objects.count(), 1)
        invitation = Invitation.objects.first()
        self.assertEqual(invitation.sender, self.assembly_manager)
        self.assertEqual(invitation.type, Invitation.InvitationType.HABITAT)
        self.assertRedirects(response, reverse('backoffice:invitation', kwargs={'pk': invitation.id}))

        # Accept invitation
        self.client.force_login(self.habitat_manager)
        self.assertIsNone(self.assembly.parent)
        activity_log_count = ActivityLogEntry.objects.count()
        response = self.client.post(
            reverse('backoffice:invitation', kwargs={'pk': invitation.id}),
            data={
                'action': 'accept',
            },
        )
        self.assertRedirects(response, reverse('backoffice:invitation', kwargs={'pk': invitation.id}))
        self.assembly.refresh_from_db()
        invitation.refresh_from_db()

        self.assertEqual(invitation.decision_by, self.habitat_manager)
        self.assertEqual(invitation.state, Invitation.RequestsState.ACCEPTED)

        self.assertEqual(self.assembly.parent, self.habitat)
        self.assertEqual(ActivityLogEntry.objects.count(), activity_log_count + 2)

    def test_send_habitat_invitation_withdraw(self):
        self.client.force_login(self.habitat_manager)
        # Send invitation from habitat to assembly
        response = self.client.post(
            reverse(
                'backoffice:invitation-send-with-pk',
                kwargs={
                    'type': 'habitat',
                    'requester_id': self.habitat.id,
                },
            ),
            data={
                'requester_id': self.habitat.id,
                'requested_id': self.assembly.id,
            },
        )
        self.assertEqual(Invitation.objects.count(), 1)
        invitation = Invitation.objects.first()
        self.assertEqual(invitation.sender, self.habitat_manager)
        self.assertEqual(invitation.type, Invitation.InvitationType.HABITAT)
        self.assertRedirects(response, reverse('backoffice:invitation', kwargs={'pk': invitation.id}))

        response = self.client.post(
            reverse('backoffice:invitation', kwargs={'pk': invitation.id}),
            data={
                'action': 'withdraw',
            },
        )
        self.assertRedirects(response, reverse('backoffice:invitation', kwargs={'pk': invitation.id}))
        self.assertEqual(Invitation.objects.count(), 1)
        invitation.refresh_from_db()
        self.assertEqual(invitation.decision_by, self.habitat_manager)
        self.assertEqual(invitation.state, Invitation.RequestsState.WITHDRAWN)

        # Test if we can send the invitation again
        response = self.client.post(
            reverse(
                'backoffice:invitation-send-with-pk',
                kwargs={
                    'type': 'habitat',
                    'requester_id': self.habitat.id,
                },
            ),
            data={
                'requester_id': self.habitat.id,
                'requested_id': self.assembly.id,
            },
        )
        self.assertEqual(Invitation.objects.count(), 2)
        invitation = Invitation.objects.filter(state=Invitation.RequestsState.REQUESTED).first()
        self.assertEqual(invitation.sender, self.habitat_manager)
        self.assertEqual(invitation.type, Invitation.InvitationType.HABITAT)
        self.assertRedirects(response, reverse('backoffice:invitation', kwargs={'pk': invitation.id}))

    def test_send_habitat_invitation_reject(self):
        with patch.object(timezone, 'now', return_value=datetime(2042, 12, 27, 12, 34, 0, 0, tzinfo=UTC)):
            self.client.force_login(self.habitat_manager)
            # Send invitation from habitat to assembly
            response = self.client.post(
                reverse(
                    'backoffice:invitation-send-with-pk',
                    kwargs={
                        'type': 'habitat',
                        'requester_id': self.habitat.id,
                    },
                ),
                data={
                    'requester_id': self.habitat.id,
                    'requested_id': self.assembly.id,
                },
            )
            self.assertEqual(Invitation.objects.count(), 1)
            invitation = Invitation.objects.first()
            self.assertEqual(invitation.sender, self.habitat_manager)
            self.assertEqual(invitation.type, Invitation.InvitationType.HABITAT)
            self.assertRedirects(response, reverse('backoffice:invitation', kwargs={'pk': invitation.id}))

            self.client.force_login(self.assembly_manager)
            response = self.client.post(
                reverse('backoffice:invitation', kwargs={'pk': invitation.id}),
                data={
                    'action': 'reject',
                },
            )
            self.assertRedirects(response, reverse('backoffice:invitation', kwargs={'pk': invitation.id}))
            self.assertEqual(Invitation.objects.count(), 1)
            invitation.refresh_from_db()
            self.assertEqual(invitation.decision_by, self.assembly_manager)
            self.assertEqual(invitation.state, Invitation.RequestsState.REJECTED)

            self.client.force_login(self.habitat_manager)
            # Try to re-Send invitation from habitat to assembly
            with patch('core.models.invitation._', mocktrans):
                response = self.client.post(
                    reverse(
                        'backoffice:invitation-send-with-pk',
                        kwargs={
                            'type': 'habitat',
                            'requester_id': self.habitat.id,
                        },
                    ),
                    data={
                        'requester_id': self.habitat.id,
                        'requested_id': self.assembly.id,
                    },
                )
            self.assertTemplateUsed(response, 'backoffice/invitations/create_edit_habitat.html')
            self.assertContains(response, 'Invitation__error__rejected_timeout_translated', status_code=200)
            self.assertEqual(Invitation.objects.count(), 1)

        # After the rejection has timed out, can send invitation again
        with (
            patch.object(timezone, 'now', return_value=(datetime(2042, 12, 27, 12, 34, 0, 0, tzinfo=UTC)) + timedelta(hours=4)),
            patch('core.models.invitation.datetime') as mock_datetime,
        ):
            mock_datetime.now.return_value = datetime(2042, 12, 27, 12, 34, 0, 0, tzinfo=UTC) + timedelta(hours=4)
            response = self.client.post(
                reverse(
                    'backoffice:invitation-send-with-pk',
                    kwargs={
                        'type': 'habitat',
                        'requester_id': self.habitat.id,
                    },
                ),
                data={
                    'requester_id': self.habitat.id,
                    'requested_id': self.assembly.id,
                },
            )
            self.assertEqual(Invitation.objects.count(), 2)
            invitation_2 = Invitation.objects.exclude(id=invitation.id).first()
            self.assertEqual(invitation_2.sender, self.habitat_manager)
            self.assertEqual(invitation_2.type, Invitation.InvitationType.HABITAT)
            self.assertRedirects(response, reverse('backoffice:invitation', kwargs={'pk': invitation_2.id}))

            # Accept invitation
            self.client.force_login(self.assembly_manager)
            self.assertIsNone(self.assembly.parent)
            activity_log_count = ActivityLogEntry.objects.count()
            response = self.client.post(
                reverse('backoffice:invitation', kwargs={'pk': invitation_2.id}),
                data={
                    'action': 'accept',
                },
            )
            self.assertRedirects(response, reverse('backoffice:invitation', kwargs={'pk': invitation_2.id}))
            self.assembly.refresh_from_db()
            invitation_2.refresh_from_db()

            self.assertEqual(invitation_2.decision_by, self.assembly_manager)
            self.assertEqual(invitation_2.state, Invitation.RequestsState.ACCEPTED)

            self.assertEqual(self.assembly.parent, self.habitat)
            self.assertEqual(ActivityLogEntry.objects.count(), activity_log_count + 2)

    def test_send_habitat_invitation_update(self):
        self.client.force_login(self.habitat_manager)
        # Send invitation from habitat to assembly
        response = self.client.post(
            reverse(
                'backoffice:invitation-send-with-pk',
                kwargs={
                    'type': 'habitat',
                    'requester_id': self.habitat.id,
                },
            ),
            data={
                'requester_id': self.habitat.id,
                'requested_id': self.assembly.id,
            },
        )
        self.assertEqual(Invitation.objects.count(), 1)
        invitation = Invitation.objects.first()
        assert invitation
        self.assertEqual(invitation.sender, self.habitat_manager)
        self.assertEqual(invitation.type, Invitation.InvitationType.HABITAT)
        self.assertRedirects(response, reverse('backoffice:invitation', kwargs={'pk': invitation.id}))

        response = self.client.post(
            reverse('backoffice:invitation', kwargs={'pk': invitation.id}),
            data={
                'action': 'update',
                'requester_id': invitation.requester.id,
                'requested_id': invitation.requested.id,
                'comment': 'Updated comment',
            },
        )
        self.assertRedirects(response, reverse('backoffice:invitation', kwargs={'pk': invitation.id}))
        self.assertEqual(Invitation.objects.count(), 1)
        invitation.refresh_from_db()
        self.assertEqual(invitation.decision_by, None)
        self.assertEqual(invitation.state, Invitation.RequestsState.REQUESTED)
        self.assertEqual(invitation.comment, 'Updated comment')

    def test_send_habitat_invitation_invalid_action(self):
        self.client.force_login(self.habitat_manager)
        # Send invitation from habitat to assembly
        response = self.client.post(
            reverse(
                'backoffice:invitation-send-with-pk',
                kwargs={
                    'type': 'habitat',
                    'requester_id': self.habitat.id,
                },
            ),
            data={
                'requester_id': self.habitat.id,
                'requested_id': self.assembly.id,
            },
        )
        self.assertEqual(Invitation.objects.count(), 1)
        invitation = Invitation.objects.first()
        self.assertEqual(invitation.sender, self.habitat_manager)
        self.assertEqual(invitation.type, Invitation.InvitationType.HABITAT)
        self.assertRedirects(response, reverse('backoffice:invitation', kwargs={'pk': invitation.id}))

        self.client.force_login(self.assembly_manager)
        response = self.client.post(
            reverse('backoffice:invitation', kwargs={'pk': invitation.id}),
            data={
                'action': 'invalid',
            },
        )
        self.assertContains(response, 'Invalid action', status_code=400)

    def test_no_permission(self):
        self.client.force_login(self.habitat_manager)
        # Send invitation from habitat to assembly
        response = self.client.post(
            reverse(
                'backoffice:invitation-send-with-pk',
                kwargs={
                    'type': 'habitat',
                    'requester_id': self.habitat.id,
                },
            ),
            data={
                'requester_id': self.habitat.id,
                'requested_id': self.assembly.id,
            },
        )
        self.assertEqual(Invitation.objects.count(), 1)
        invitation = Invitation.objects.first()
        self.assertEqual(invitation.sender, self.habitat_manager)
        self.assertEqual(invitation.type, Invitation.InvitationType.HABITAT)
        self.assertRedirects(response, reverse('backoffice:invitation', kwargs={'pk': invitation.id}))

        # Cannot accept invitation
        self.assertIsNone(self.assembly.parent)
        response = self.client.post(
            reverse('backoffice:invitation', kwargs={'pk': invitation.id}),
            data={
                'action': 'accept',
            },
        )
        self.assertContains(response, 'Forbidden', status_code=403)
        self.assembly.refresh_from_db()
        self.assertIsNone(self.assembly.parent)
        invitation.refresh_from_db()
        self.assertIsNone(invitation.decision_by)
        self.assertEqual(invitation.state, Invitation.RequestsState.REQUESTED)

        # Cannot reject invitation
        self.assertIsNone(self.assembly.parent)
        response = self.client.post(
            reverse('backoffice:invitation', kwargs={'pk': invitation.id}),
            data={
                'action': 'reject',
            },
        )
        self.assertContains(response, 'Forbidden', status_code=403)
        self.assembly.refresh_from_db()
        self.assertIsNone(self.assembly.parent)
        invitation.refresh_from_db()
        self.assertIsNone(invitation.decision_by)
        self.assertEqual(invitation.state, Invitation.RequestsState.REQUESTED)

        self.client.force_login(self.assembly_manager)
        # Cannot withdraw invitation
        self.assertIsNone(self.assembly.parent)
        response = self.client.post(
            reverse('backoffice:invitation', kwargs={'pk': invitation.id}),
            data={
                'action': 'withdraw',
            },
        )
        self.assertContains(response, 'Forbidden', status_code=403)
        self.assembly.refresh_from_db()
        self.assertIsNone(self.assembly.parent)
        invitation.refresh_from_db()
        self.assertIsNone(invitation.decision_by)
        self.assertEqual(invitation.state, Invitation.RequestsState.REQUESTED)

        # Cannot update invitation
        self.assertIsNone(self.assembly.parent)
        response = self.client.post(
            reverse('backoffice:invitation', kwargs={'pk': invitation.id}),
            data={
                'action': 'update',
                'requester_id': invitation.requester.id,
                'requested_id': invitation.requested.id,
                'comment': 'Updated comment',
            },
        )
        self.assertContains(response, 'Forbidden', status_code=403)
        self.assembly.refresh_from_db()
        self.assertIsNone(self.assembly.parent)
        invitation.refresh_from_db()
        self.assertIsNone(invitation.decision_by)
        self.assertEqual(invitation.state, Invitation.RequestsState.REQUESTED)
        self.assertEqual(invitation.comment, '')

    def test_duplicate_invitation(self):
        self.client.force_login(self.habitat_manager)
        # Send invitation from habitat to assembly
        response = self.client.post(
            reverse(
                'backoffice:invitation-send-with-pk',
                kwargs={
                    'type': 'habitat',
                    'requester_id': self.habitat.id,
                },
            ),
            data={
                'requester_id': self.habitat.id,
                'requested_id': self.assembly.id,
            },
        )
        self.assertEqual(Invitation.objects.count(), 1)
        invitation = Invitation.objects.first()
        self.assertEqual(invitation.sender, self.habitat_manager)
        self.assertEqual(invitation.type, Invitation.InvitationType.HABITAT)
        self.assertRedirects(response, reverse('backoffice:invitation', kwargs={'pk': invitation.id}))

        # Try to send duplicate invitation

        # Send invitation from habitat to assembly
        with patch('core.models.invitation._', mocktrans):
            response = self.client.post(
                reverse(
                    'backoffice:invitation-send-with-pk',
                    kwargs={
                        'type': 'habitat',
                        'requester_id': self.habitat.id,
                    },
                ),
                data={
                    'requester_id': self.habitat.id,
                    'requested_id': self.assembly.id,
                },
            )
        self.assertTemplateUsed(response, 'backoffice/invitations/create_edit_habitat.html')
        self.assertContains(response, f'Invitation__error__already_exists {self.habitat.name} {self.assembly.name} {self.habitat_manager.username}_translated')
        self.assertEqual(Invitation.objects.count(), 1)
