from typing import Any

from django.contrib.auth.views import RedirectURLMixin
from django.contrib.messages import SUCCESS, add_message
from django.http import HttpRequest
from django.http.response import HttpResponse
from django.urls import reverse
from django.views.generic import CreateView, UpdateView, View
from django.views.generic.edit import ModelFormMixin

from core.forms import InvitationHabitatForm
from core.models import Invitation

from backoffice.views.invitations.mixins import HandleInvitationActionMixin
from backoffice.views.mixins import ConferenceLoginRequiredMixin


class InvitationHabitatBaseView(ConferenceLoginRequiredMixin, RedirectURLMixin, ModelFormMixin, View):
    template_name = 'backoffice/invitations/create_edit_habitat.html'
    form_class = InvitationHabitatForm
    model = Invitation

    def get_form_kwargs(self) -> dict[str, Any]:
        return {
            **super().get_form_kwargs(),
            'conference': self.conference,
            'user': self.request.user,
            'requester_id': self.kwargs.get('requester_id'),
        }

    def get_default_redirect_url(self) -> str:
        if not hasattr(self, 'object'):  # pragma: no cover
            return reverse('backoffice:invitations')
        return reverse('backoffice:invitation', kwargs={'pk': self.object.pk})


class InvitationHabitatSendView(InvitationHabitatBaseView, CreateView):
    def dispatch(self, request: HttpRequest, *args: Any, **kwargs: Any) -> HttpResponse:
        if kwargs.get('requester_id') is None:
            return HttpResponse('Requester ID is required', status=400)
        return super().dispatch(request, *args, **kwargs)

    def form_valid(self, form):
        response = super().form_valid(form)
        add_message(self.request, SUCCESS, f'Invitation to {self.object.requested.name} sent')
        return response


class InvitationHabitatUpdateView(HandleInvitationActionMixin, InvitationHabitatBaseView, UpdateView):
    object: Invitation

    def has_permission(self):
        invitation = self.get_object()
        if not Invitation.type_is(invitation):  # pragma: no cover
            raise ValueError('Invalid invitation type')
        return self.request.user.has_perm('core.change_assembly', invitation.requested) or self.request.user.has_perm(
            'core.change_assembly', invitation.requester
        )

    def post(self, request: HttpRequest, *args: str, **kwargs: Any) -> HttpResponse:
        self.object = self.get_object()  # pylint: disable=attribute-defined-outside-init
        if not Invitation.type_is(self.object):  # pragma: no cover
            raise ValueError('Invalid invitation type')
        action = request.POST.get('action', '')
        if action in ['update', 'withdraw'] and not request.user.has_perm('core.change_assembly', self.object.requester):
            return self.handle_no_permission()
        if action in ['accept', 'reject'] and not request.user.has_perm('core.change_assembly', self.object.requested):
            return self.handle_no_permission()
        return super().post(request, *args, **kwargs)
