from typing import Any

from django.contrib.messages import ERROR, SUCCESS, WARNING, add_message
from django.http import HttpRequest, HttpResponseRedirect
from django.http.response import HttpResponse
from django.utils.translation import gettext_lazy as _
from django.views.generic import UpdateView

from core.models import Invitation


class HandleInvitationActionMixin(UpdateView):
    object: Invitation

    def post(self, request: HttpRequest, *args: str, **kwargs: Any) -> HttpResponse:
        action = request.POST.get('action', '')
        if not hasattr(self, 'object') and not Invitation.type_is(self.object):  # pragma: no cover
            raise ValueError('Invalid invitation type')
        match action:
            case 'update':
                response = super().post(request, *args, **kwargs)
                add_message(
                    request,
                    SUCCESS,
                    _('Invitation__updated__message %(requester)s %(requested)s')
                    % {
                        'requester': self.object.requester.name,
                        'requested': self.object.requested.name,
                    },
                )
                return response
            case 'withdraw':
                self.object.withdraw(request.user)
                add_message(
                    request,
                    WARNING,
                    _('Invitation__withdraw__message %(requester)s %(requested)s')
                    % {
                        'requester': self.object.requester.name,
                        'requested': self.object.requested.name,
                    },
                )
                return HttpResponseRedirect(self.get_success_url())
            case 'accept':
                self.object.accept(request.user)
                add_message(
                    request,
                    SUCCESS,
                    _('Invitation__accept__message %(requester)s %(requested)s')
                    % {
                        'requester': self.object.requester.name,
                        'requested': self.object.requested.name,
                    },
                )
                return HttpResponseRedirect(self.get_success_url())
            case 'reject':
                self.object.reject(request.user)
                add_message(
                    request,
                    ERROR,
                    _('Invitation__rejected__message %(requester)s %(requested)s')
                    % {
                        'requester': self.object.requester.name,
                        'requested': self.object.requested.name,
                    },
                )
                return HttpResponseRedirect(self.get_success_url())
            case _:
                return HttpResponse('Invalid action', status=400)
