import logging

from oauth2_provider.models import AccessToken

from django.contrib import messages
from django.contrib.sessions.exceptions import SuspiciousSession
from django.contrib.sessions.models import Session
from django.http import HttpRequest
from django.shortcuts import redirect
from django.views.generic import DetailView, ListView

from core.models import BulletinBoardEntry, DirectMessage, PlatformUser

from backoffice.views.moderation.mixins import ModerationAdminMixin

logger = logging.getLogger(__name__)


class ModerationUserListView(ModerationAdminMixin, ListView):
    template_name = 'backoffice/moderation_user-list.html'

    def get_queryset(self):
        qs = PlatformUser.objects.filter(conferences__conference=self.conference)
        return qs


class ModerationUserDetailView(ModerationAdminMixin, DetailView):
    template_name = 'backoffice/moderation_user-detail.html'

    def get_queryset(self):
        qs = PlatformUser.objects.filter(conferences__conference=self.conference)
        return qs

    def get_context_data(self, *args, **kwargs):
        ctx = super().get_context_data(*args, **kwargs)
        ctx['user_cm'] = self.conference.users.filter(user=ctx['object']).first()
        return ctx

    def block_user(self, request: HttpRequest, user: PlatformUser):
        user.is_active = False
        user.save(update_fields=['is_active'])
        messages.success(request, f'user #{user.id} "{user.username}" set non-active (i.e. blocked)')
        logger.warning('User "%s" was disabled by %s.', user.username, request.user)

        log = []
        deleted_access_tokens = 0
        for token in AccessToken.objects.filter(user_id=user.id):
            log.append(str(token.application))
            token.delete()
            deleted_access_tokens += 1

        if deleted_access_tokens > 0:
            messages.info(self.request, f'OAuth2 access token by {user.username} deleted for application(s):\n' + ',\n'.join(log))
        else:
            messages.info(self.request, f'No OAuth2 access tokens by {user.username} to be deleted.')

        deleted_sessions = 0
        for session in Session.objects.all():
            try:
                session_data = session.get_decoded()
                if session_data.get('_auth_user_id') == str(user.pk) and session_data.get('_auth_user_backend') == 'django.contrib.auth.backends.ModelBackend':
                    session.delete()
                    deleted_sessions += 1

            except SuspiciousSession:
                # while we are at it, delete the errornous session, too
                session.delete()

        if deleted_sessions > 0:
            messages.info(self.request, f'Deleted {deleted_sessions} sessions.')
        else:
            messages.info(self.request, 'No sessions to delete o.O')

    def post(self, request, *args, **kwargs):
        user = self.get_object()
        action = request.POST.get('action')

        # TODO: add log entries
        if action == 'block':
            self.block_user(request, user)

        elif action == 'unblock':
            if user.is_active:
                messages.warning(self.request, 'User was already enabled: ' + user.username)

            else:
                user.is_active = True
                user.save(update_fields=['is_active'])
                messages.success(self.request, 'User re-enabled: ' + user.username)
                logger.warning('User "%s" was re-enabled by %s.', user.username, self.request.user)

        elif action == 'ban':
            user.shadow_banned = True
            user.save(update_fields=['shadow_banned'])

            DirectMessage.objects.filter(sender=user).exclude(recipient=user).update(deleted_by_recipient=True)
            BulletinBoardEntry.objects.filter(owner=user).update(hidden=True)

            messages.success(self.request, 'User shadowbanned: ' + user.username)
            logger.warning('User "%s" was shadowbanned by %s.', user.username, self.request.user)

        elif action == 'unban':
            user.shadow_banned = False
            user.save(update_fields=['shadow_banned'])

            messages.success(self.request, 'User unbanned: ' + user.username)
            logger.warning('User "%s" shadowban removed by %s.', user.username, self.request.user)

        else:
            messages.error(request, f'UNKNOWN ACTION: {action}')

        return redirect('backoffice:moderation-user-detail', pk=user.pk)


class ModerationUserRenameView(ModerationAdminMixin, DetailView):
    template_name = 'backoffice/moderation_user-rename.html'

    def get_queryset(self):
        qs = PlatformUser.objects.filter(conferences__conference=self.conference)
        return qs

    def post(self, *args, **kwargs):
        user = self.get_object()

        new_name = self.request.POST.get('new_name', '').strip()
        if new_name in ('', user.username):
            return self.get(*args, **kwargs)

        if PlatformUser.objects.filter(username=new_name).exists():
            messages.error(self.request, f'Username "{new_name}" exists, won\'t rename "{user.username}"!')
            return self.get(*args, **kwargs)

        old_name = user.username
        user.username = new_name
        user.generate_slug()
        user.save()

        messages.success(self.request, f'Renamed user "{old_name}" to "{new_name}".')
        logger.warning('Renamed user "%s" to "%s" (by %s).', old_name, new_name, self.request.user.username)

        return redirect('backoffice:moderation-user-detail', pk=user.pk)
