from django.contrib.contenttypes.models import ContentType
from django.db.models import QuerySet
from django.forms import ModelForm, Select
from django.utils.translation import gettext_lazy as _

from core.models import Assembly, Conference, Invitation, PlatformUser


class InvitationBaseForm(ModelForm):
    class Meta:
        model = Invitation
        fields = (
            'requester_id',
            'requested_id',
            'comment',
        )

    requester_id: str | None = None

    def __init__(self, *args, conference: Conference, user: PlatformUser, instance: Invitation | None, requester_id: str | None = None, **kwargs):
        self.create = instance is None
        self.conference = conference
        self.user = user
        if requester_id is not None:
            self.requester_id = str(requester_id)
            kwargs['initial'] = {
                'requester_id': requester_id,
            }
            if 'data' in kwargs:
                kwargs['data'] = kwargs['data'].copy()
                kwargs['data']['requester_id'] = requester_id
        super().__init__(*args, instance=instance, **kwargs)
        if instance and instance.state != Invitation.RequestsState.REQUESTED:
            del self.fields['comment']
        if requester_id is not None:
            self.fields['requester_id'].widget.attrs['disabled'] = True

    def save(self, commit: bool = True) -> Invitation:
        invitation = super().save(commit=False)
        if self.create:
            invitation.conference = self.conference
            invitation.sender = self.user
        if commit:
            invitation.save()
        return invitation


class InvitationHabitatForm(InvitationBaseForm):
    class Meta(InvitationBaseForm.Meta):
        widgets = {
            'requester_id': Select(),
            'requested_id': Select(),
        }

    requester_id: str | None = None

    def get_queryset(self) -> QuerySet[Assembly]:
        return Assembly.objects.associated_with_user(conference=self.conference, user=self.user, show_public=True, staff_can_see=False)

    def get_choices(self, *, habitat: bool) -> list[tuple[str, str]]:
        if habitat:
            return self.get_queryset().exclude(hierarchy=Assembly.Hierarchy.REGULAR).values_list('id', 'name')
        else:
            return self.get_queryset().filter(hierarchy=Assembly.Hierarchy.REGULAR).values_list('id', 'name')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.create:
            if not self.requester_id:
                raise ValueError('Requester ID is required')
            try:
                assembly = self.get_queryset().get(id=self.requester_id)
            except Assembly.DoesNotExist as exc:
                raise ValueError('Invalid requester ID') from exc
            habitat = assembly.hierarchy != Assembly.Hierarchy.REGULAR
            self.fields['requester_id'].label = _('Assembly__parent') if habitat else _('Assembly')
            self.fields['requested_id'].label = _('Assembly') if habitat else _('Assembly__parent')
            self.fields['requester_id'].widget.choices = self.get_choices(habitat=habitat)
            self.fields['requested_id'].widget.choices = self.get_choices(habitat=(not habitat))
        else:
            self.permissions = {
                'requester': self.user.has_perm('core.change_assembly', self.instance.requester),
                'requested': self.user.has_perm('core.change_assembly', self.instance.requested),
            }
            if not self.permissions['requester'] and 'comment' in self.fields:
                del self.fields['comment']
            habitat = self.instance.requester.hierarchy != Assembly.Hierarchy.REGULAR
        self.requester_label = _('Assembly__parent') if habitat else _('Assembly')
        self.requested_label = _('Assembly__parent') if not habitat else _('Assembly')

    def save(self, commit: bool = True) -> Invitation:
        invitation = super().save(commit=False)
        assembly_type = ContentType.objects.get_for_model(Assembly)
        invitation.type = Invitation.InvitationType.HABITAT
        invitation.requester_type = assembly_type
        invitation.requested_type = assembly_type
        if commit:
            invitation.save()
        return invitation
