from typing import Any

from django.core.exceptions import ValidationError
from django.forms import CharField
from django.utils.translation import gettext_lazy as _

from core.base_forms import TranslatedFieldsForm
from core.models import Assembly, PlatformUser, Project
from core.models.tags import clean_tags


class ProjectForm(TranslatedFieldsForm):
    class Meta:
        model = Project
        fields = (
            'name',
            'description',
            'banner_image',
            'location',
            'is_public',
            'tags_list',
        )

    def save(self, commit: bool = True) -> Any:
        project: Project = super().save(False)
        project.update_tags(self.cleaned_data['tags_list'], auto_create=True)
        if commit:
            project.save()
        return project

    tags_list = CharField(required=False, label=_('Tags'), help_text=_('Project__tags__help'))

    def __init__(
        self,
        *args,
        conference,
        assembly: Assembly | None = None,
        owner: PlatformUser | None = None,
        instance: Project | None = None,
        publish: bool = False,
        **kwargs,
    ):
        self.conference = conference
        self.assembly = assembly
        self.owner = owner
        self.publish = publish
        self.create = instance is None
        super().__init__(*args, instance=instance, **kwargs)
        if self.instance.pk:
            self.initial['tags_list'] = ', '.join(self.instance.sorted_tags)

    def clean_name(self):
        if (
            self.instance.assembly
            and Project.objects.filter(assembly=self.instance.assembly, name=self.cleaned_data['name']).exclude(pk=self.instance.pk).exists()
        ):
            raise ValidationError(_('Project__name__assembly_unique'))
        if self.instance.assembly and Project.objects.filter(owner=self.instance.owner, name=self.cleaned_data['name']).exclude(pk=self.instance.pk).exists():
            raise ValidationError(_('Project__name__owner_unique'))

        # update slug to be based on the new name iff the name changed.
        if self.instance.name != self.cleaned_data['name']:
            self.instance.name = self.cleaned_data['name']
            self.instance.generate_slug()
            self.cleaned_data['slug'] = self.instance.slug
        return self.cleaned_data['name']

    def clean_tags_list(self):
        return clean_tags(self.cleaned_data['tags_list'])

    def clean(self) -> dict[str, Any]:
        if not self.create:
            if self.instance.assembly and self.assembly and self.instance.assembly != self.assembly:
                raise ValidationError(_('Project__assembly__unchangeable'))
            if self.instance.owner and self.instance.owner != self.owner:
                raise ValidationError(_('Project__owner__unchangeable'))
            if self.instance.conference and self.instance.conference != self.conference:
                raise ValidationError(_('Project__conference__unchangeable'))
        self.cleaned_data['is_public'] = not self.instance.is_public if self.publish else self.instance.is_public
        self.instance.conference = self.conference
        self.instance.assembly = self.assembly
        self.instance.owner = self.owner
        return super().clean()
