import io
import logging
from pathlib import Path
from string import ascii_lowercase, digits
from typing import TypeIs
from uuid import uuid4

from segno import make as segno_make

from django.conf import settings
from django.contrib.postgres.fields import DateTimeRangeField
from django.core.exceptions import ValidationError
from django.core.files.base import ContentFile
from django.core.files.storage import default_storage
from django.core.validators import RegexValidator
from django.db import models
from django.db.models import Q, QuerySet
from django.template.loader import render_to_string
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from core.choices import UserBadgeVisibility
from core.fields import ConferenceReference
from core.markdown import compile_translated_markdown_fields, store_relationships
from core.models.assemblies import Assembly
from core.models.base_managers import ConferenceManagerMixin
from core.models.conference import Conference, ConferenceMember
from core.models.users import PlatformUser
from core.utils import generate_token
from core.validators import ImageDimensionValidator

logger = logging.getLogger(__name__)

TOKEN_VALIDATOR = RegexValidator(r'^[0-9a-z]*$', _('BadgeToken__validator_info__lower_alpha_num'))


class BadgeCategory(models.Model):
    class Meta:
        unique_together = [['conference', 'slug']]
        verbose_name = _('BadgeCategory')
        verbose_name_plural = _('BadgeCategories')

    conference = ConferenceReference(related_name='badge_categories')
    name = models.CharField(
        max_length=50,
        help_text=_('BadgeCategory__name__help'),
        verbose_name=_('BadgeCategory__name'),
        unique=True,
    )
    slug = models.SlugField(help_text=_('BadgeCategory__slug__help'), verbose_name=_('BadgeCategory__slug'))
    symbol = models.SlugField(
        max_length=50,
        help_text=_('BadgeCategory__symbol__help'),
        verbose_name=_('BadgeCategory__symbol'),
        blank=True,
    )
    description = models.TextField()

    def __str__(self):
        return self.name


def get_badge_filename(instance: 'Badge', filename: str):
    return Path('badges').joinpath(str(instance.id), filename)


class BadgeManager(ConferenceManagerMixin['Badge']):
    def apply_public_filter(self, queryset: 'QuerySet[Badge]', member: ConferenceMember | None = None) -> 'QuerySet[Badge]':
        if member is None:
            return queryset.filter(state=Badge.State.PUBLIC)
        return queryset.filter(Q(state=Badge.State.PUBLIC) | Q(users__user=member.user))

    def accessible_by_user(self, user: PlatformUser, conference: Conference, staff_can_manage=True) -> 'QuerySet[Badge]':
        if user is None or not user.is_authenticated:
            user = PlatformUser.get_anonymous_user()
        qs = self.get_queryset()
        if not user.is_authenticated:
            return qs.filter(state=Badge.State.PUBLIC)

        manageable = Assembly.objects.manageable_by_user(conference, user=user, staff_can_manage=staff_can_manage)
        return qs.filter(Q(state=Badge.State.PUBLIC) | Q(users__user=user) | Q(issuing_assembly__in=manageable))


def get_badge_image_help_text() -> str:
    return _('Badge__image__help %(min_width)d, %(min_height)d, %(max_width)d, %(max_height)d') % {
        'min_width': settings.BADGE_IMAGE_WIDTH_MINIMUM,
        'min_height': settings.BADGE_IMAGE_HEIGHT_MINIMUM,
        'max_width': settings.BADGE_IMAGE_WIDTH_MAXIMUM,
        'max_height': settings.BADGE_IMAGE_HEIGHT_MAXIMUM,
    }


class Badge(models.Model):
    class Meta:
        unique_together = [['name', 'issuing_assembly']]

    class State(models.TextChoices):
        PLANNED = 'planned', _('Badge__state__planned')
        HIDDEN = 'hidden', _('Badge__state__hidden')
        PUBLIC = 'public', _('Badge__state__public')

    objects = BadgeManager()

    id = models.UUIDField(default=uuid4, primary_key=True, editable=False, serialize=False)
    conference = ConferenceReference(related_name='badges')
    issuing_assembly = models.ForeignKey(Assembly, related_name='badges', on_delete=models.CASCADE)

    name = models.CharField(
        max_length=100,
        help_text=_('Badge__name__help'),
        verbose_name=_('Badge__name'),
    )

    state = models.CharField(
        max_length=20,
        default=State.PLANNED,
        choices=State.choices,
        verbose_name=_('Badge__state'),
        help_text=_('Badge__state__help'),
    )
    category = models.ForeignKey(
        BadgeCategory,
        null=True,
        help_text=_('Badge__category__help'),
        verbose_name=_('Badge__category'),
        on_delete=models.SET_NULL,
    )
    symbol = models.SlugField(
        max_length=50,
        help_text=_('Badge__symbol__help'),
        verbose_name=_('Badge__symbol'),
        blank=True,
        default='',
    )

    description = models.TextField(
        blank=True,
        help_text=_('Badge__description__help'),
        verbose_name=_('Badge__description'),
    )
    description_html = models.TextField(blank=True, null=True)

    location = models.CharField(
        blank=True,
        help_text=_('Badge__location__help'),
        verbose_name=_('Badge__location'),
    )

    image_height = models.PositiveIntegerField(blank=True, null=True)
    image_width = models.PositiveIntegerField(blank=True, null=True)
    image = models.ImageField(
        blank=True,
        null=True,
        height_field='image_height',
        width_field='image_width',
        help_text=get_badge_image_help_text,
        upload_to=get_badge_filename,
        verbose_name=_('Badge__image'),
        validators=[
            ImageDimensionValidator(
                min_size=(
                    settings.BADGE_IMAGE_WIDTH_MINIMUM,
                    settings.BADGE_IMAGE_HEIGHT_MINIMUM,
                ),
                max_size=(
                    settings.BADGE_IMAGE_WIDTH_MAXIMUM,
                    settings.BADGE_IMAGE_HEIGHT_MAXIMUM,
                ),
                square=True,
            )
        ],
    )

    issuing_token = models.CharField(max_length=100, blank=True)

    def __str__(self):
        return self.name

    @classmethod
    def type_is(cls, obj: object) -> TypeIs['Badge']:
        return isinstance(obj, cls)

    @property
    def is_public(self):
        return self.state == self.State.PUBLIC

    def save(self, *args, update_fields=None, **kwargs):
        if not hasattr(self, 'conference') and hasattr(self, 'issuing_assembly'):
            self.conference = self.issuing_assembly.conference
        elif hasattr(self, 'conference') and self.conference != self.issuing_assembly.conference:
            raise ValidationError(_('Badge__issuing_assembly__wrong_conference'))
        if update_fields is None or 'description' in update_fields:
            render_results = compile_translated_markdown_fields(self, self.conference, 'description')
            store_relationships(self.conference, self, render_results)

        return super().save(*args, **kwargs)

    def reset_token(self):
        self.issuing_token = token = generate_token()
        self.save()
        return token

    def get_symbol(self):
        """Returns a symbol for this badge - either the directly assigned one or the one of the badge's category."""
        return self.symbol or self.category.symbol


def create_new_token():
    not_unique = True
    while not_unique:
        token = generate_token(20, [*ascii_lowercase, *digits])
        if not BadgeToken.objects.filter(token=token).exists():
            not_unique = False
    return str(token)


class TokenInvalid(Exception):
    msg: str

    def __init__(self, message, *args: object) -> None:
        self.msg = message
        super().__init__(*args)


class BadgeToken(models.Model):
    id = models.UUIDField(default=uuid4, primary_key=True, editable=False, serialize=False)
    badge = models.ForeignKey(Badge, related_name='tokens', on_delete=models.CASCADE)
    token = models.CharField(
        editable=False,
        unique=True,
        default=create_new_token,
        max_length=20,
        validators=[TOKEN_VALIDATOR],
    )
    issued = models.DateTimeField(auto_now_add=True)
    active = models.BooleanField(
        default=True,
        help_text=_('BadgeToken__active__help'),
        verbose_name=_('BadgeToken__active'),
    )
    comment = models.CharField(
        blank=True,
        max_length=50,
        help_text=_('BadgeToken__comment__help'),
        verbose_name=_('comment'),
    )

    redeemable_count = models.IntegerField(
        default=0,
        help_text=_('BadgeToken__redeemable_count__help'),
        verbose_name=_('BadgeToken__redeemable_count'),
    )
    qr_full = models.ImageField(
        null=True,
        help_text=_(_('BadgeToken__QRCode__help')),
        verbose_name=_('BadgeToken__QRCode'),
    )

    @property
    def redeemed_count(self):
        return self.users.count()

    @property
    def valid(self):
        try:
            self.check_validity()
        except TokenInvalid:
            return False
        return True

    @property
    def has_qr_code(self):
        return self.qr_full and default_storage.exists(self.qr_full.name)

    def save(self, *args, **kwargs) -> None:
        if not self.has_qr_code:
            from core.templatetags.hub_absolute import hub_absolute  # pylint: disable=import-outside-toplevel

            qr_full = segno_make(f"{hub_absolute('plainui:manage_badges')}?redeem_token={self.token}", micro=False)
            buffer = io.BytesIO()
            qr_full.save(buffer, kind='svg', scale=3, dark='#555')
            buffer_value = buffer.getvalue().decode('utf-8').replace('#555', 'currentColor')
            self.qr_full.save(
                f'BadgeToken_{self.badge.id}_{self.token}.svg',
                ContentFile(buffer_value),
                save=False,
            )
            logger.debug('Generated qr_code for BadgeToken %s', self.id)
        return super().save(*args, **kwargs)

    def check_validity(self):
        if self.redeemable_count and self.redeemable_count <= self.users.count():
            raise TokenInvalid(_('BadgeToken__error__no_more_redemptions %(badge)s %(badge_token)s') % {'badge': self.badge, 'badge_token': self.token})
        if not self.active:
            raise TokenInvalid(_('BadgeToken__error__inactive %(badge)s %(badge_token)s') % {'badge': self.badge, 'badge_token': self.token})
        if self.time_constraints.exists():
            now = timezone.now()
            if not self.time_constraints.filter(date_time_range__contains=now):
                raise TokenInvalid(_('BadgeToken__error__currently_inactive %(badge)s %(badge_token)s') % {'badge': self.badge, 'badge_token': self.token})
        return True

    def __str__(self):
        return f'{self.badge}: {self.token}'

    @classmethod
    def type_is(cls, obj: object) -> 'TypeIs[BadgeToken]':
        return isinstance(obj, cls)


class BadgeTokenTimeConstraint(models.Model):
    badge_token = models.ForeignKey(to=BadgeToken, related_name='time_constraints', on_delete=models.CASCADE, editable=False)
    date_time_range = DateTimeRangeField(
        help_text=_('BadgeTokenTimeConstraint__date_time_range__help'),
        verbose_name=_('BadgeTokenTimeConstraint__date_time_range'),
    )


class UserBadgeManager(models.Manager):
    def redeem_badge(
        self,
        user: PlatformUser,
        badge: Badge | None = None,
        token: BadgeToken | None = None,
        issuer: PlatformUser | None = None,
    ):
        if badge and token:
            raise ValueError("Both 'badge' and 'token' were supplied")
        elif not badge and not token:
            raise ValueError("Neither 'badge' and 'token' were supplied")
        elif token:
            badge = token.badge

        if token:
            token.check_validity()
        assert user is not None
        user_badge, created = self.get_or_create(
            user=user,
            badge=badge,
            defaults={
                'accepted_by_user': not bool(issuer),
                'token': token if token else None,
                'visibility': user.conferences.filter(conference=badge.conference).first().default_badge_visibility,
            },
        )

        # If it isn't redeemed by the user himself, then he will be notified and needs to accept it
        if issuer and created:
            ctx = {
                'badge': badge,
                'issuer': f'Badge Token Redeem ({token})' if token else f'Assembly {badge.issuing_assembly}{f" - {issuer}" if issuer else ""}',
            }

            notify_ok = user.notify_user(
                # TODO: Subject+Message in der Sprache des users
                subject=f'Badge: "{badge.name}"',
                message_text=render_to_string('core/badge_awarded.txt', ctx),
            )

            if not notify_ok:
                logger.warning('User %s could not be notified about the awarded badge %s.', user, self)

        if token:
            logger.info(
                'User "%(user)s" redeemed the token "%(token)s" for badge "%(badge)s". Nr.: %(count)s%(redeemable_count)s',
                {
                    'user': user,
                    'token': token.token,
                    'badge': badge,
                    'count': token.users.count(),
                    'redeemable_count': '/{token.redeemable_count}' if token.redeemable_count else '',
                },
            )
        else:
            logger.info(
                'Awarded badge "%(badge)s" to user "%(user)s" (Issuer: %(issuer)s). Nr.: %(count)s',
                {
                    'badge': badge,
                    'user': user,
                    'issuer': issuer,
                    'count': badge.users.count(),
                },
            )

        return (user_badge, created)


class UserBadge(models.Model):
    objects = UserBadgeManager()

    class Meta:
        constraints = [
            models.constraints.UniqueConstraint(fields=['user', 'badge'], name='unique_userbadge_user_badge'),
        ]

    user = models.ForeignKey(PlatformUser, related_name='badges', on_delete=models.CASCADE)
    badge = models.ForeignKey(Badge, related_name='users', on_delete=models.CASCADE)
    token = models.ForeignKey(BadgeToken, related_name='users', on_delete=models.CASCADE, null=True)

    accepted_by_user = models.BooleanField(default=False)
    visibility = models.CharField(
        max_length=20,
        default=UserBadgeVisibility.PRIVATE,
        choices=UserBadgeVisibility.choices,
        verbose_name=_('UserBadge__visibility'),
        help_text=_('UserBadge__visibility__help'),
    )

    def __str__(self):
        return f'{self.user}: {self.badge}'

    def acceptByUser(self):
        self.accepted_by_user = True
        self.save()

    def revokeAcceptByUser(self):
        self.accepted_by_user = False
        self.save()

    def changeVisibiltyState(self, state: UserBadgeVisibility):
        self.visibility = state
        self.save()
