from datetime import datetime, timedelta
from typing import TYPE_CHECKING, TypeIs
from uuid import uuid4

from django.conf import settings
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from django.db import models, transaction
from django.db.models import Max
from django.utils.translation import gettext_lazy as _

from core.fields import ConferenceReference
from core.models.activitylog import ActivityLogChange

if TYPE_CHECKING:  # pragma: no cover
    from core.models.users import PlatformUser


class InvitationAlreadyExists(ValidationError):
    """
    This exception is raised when an invitation already exists between a requester and a requested entity.
    """

    def __init__(self, invitation: 'Invitation') -> None:
        self.invitation = invitation
        super().__init__(
            _('Invitation__error__already_exists %(requester)s %(requested)s %(sender)s'),
            params={'requester': invitation.requester, 'requested': invitation.requested, 'sender': invitation.sender},
            code='already_exists',
        )


class InvitationRejectedTimeout(ValidationError):
    """
    This exception is raised when an invitation has been rejected and the timeout has not yet expired.
    """

    def __init__(self) -> None:
        super().__init__(
            _('Invitation__error__rejected_timeout'),
            code='rejected_timeout',
        )


class RequestsState(models.TextChoices):
    """
    Enumeration for the state of requests.

    Attributes:
        REQUESTED: The request has been made.
        WITHDRAWN: The request has been withdrawn.
        ACCEPTED: The request has been accepted.
        REJECTED: The request has been rejected.
    """

    REQUESTED = 'requested', _('Invitation__requests_state-requested')
    WITHDRAWN = 'withdrawn', _('Invitation__requests_state-withdrawn')
    ACCEPTED = 'accepted', _('Invitation__requests_state-accepted')
    REJECTED = 'rejected', _('Invitation__requests_state-rejected')


class InvitationType(models.TextChoices):
    """
    Enumeration for the type of invitations.

    Attributes:
        ASSEMBLY_MEMBER: Invitation for an Conferencemember to join an assembly.
        TEAM_MEMBER: Invitation for an Conferencemember to join a team.
        HABITAT: Invitation of ah Assembly for a habitat.
    """

    ASSEMBLY_MEMBER = 'ASSEMBLY_MEMBER', _('Assembly Member')
    TEAM_MEMBER = 'TEAM_MEMBER', _('Team Member')
    HABITAT = 'HABITAT', _('Habitat')


class Invitation(models.Model):
    """
    Model representing an invitation.

    Attributes:
        requester_type: The type of the requester (ContentType).
        requester_id: The ID of the requester.
        requester: The requester (GenericForeignKey).
        requested_type: The type of the requested entity (ContentType).
        requested_id: The ID of the requested entity.
        requested: The requested entity (GenericForeignKey).
        comment: A comment associated with the invitation.
    """

    class Meta:
        indexes = [
            models.Index(fields=['requester_id', 'requested_id']),
            models.Index(fields=['requester_type', 'requester_id'], name='requester_idx'),
            models.Index(fields=['requested_type', 'requested_id'], name='requested_idx'),
        ]
        verbose_name = _('Invitation__name')
        verbose_name_plural = _('Invitation__name__plural')

    # Make the enums available through the class
    InvitationType = InvitationType
    RequestsState = RequestsState

    id = models.UUIDField(primary_key=True, default=uuid4, editable=False)
    conference = ConferenceReference(related_name='+')
    requester_type = models.ForeignKey(ContentType, on_delete=models.CASCADE, related_name='invitations_sent')
    requester_id = models.CharField(max_length=36)
    requester = GenericForeignKey('requester_type', 'requester_id')
    requested_type = models.ForeignKey(ContentType, on_delete=models.CASCADE, related_name='invitations_received')
    requested_id = models.CharField(max_length=36)
    requested = GenericForeignKey('requested_type', 'requested_id')

    comment = models.TextField(blank=True, verbose_name=_('Invitation__comment'), help_text=_('Invitation__comment__help_text'))

    requested_at = models.DateTimeField(auto_now_add=True, editable=False, verbose_name=_('Invitation__requested_at'))
    updated_at = models.DateTimeField(auto_now=True, editable=False, verbose_name=_('Invitation__updated_at'))
    sender = models.ForeignKey(
        'PlatformUser',
        on_delete=models.SET_NULL,
        verbose_name=_('Invitation__sender'),
        null=True,
    )
    decision_by = models.ForeignKey(
        'PlatformUser',
        on_delete=models.SET_DEFAULT,
        verbose_name=_('Invitation__decision_by'),
        null=True,
        default=None,
        related_name='+',
    )

    type = models.CharField(
        max_length=20,
        choices=InvitationType.choices,
        verbose_name=_('Invitation__type'),
    )
    state = models.CharField(
        max_length=20,
        choices=RequestsState.choices,
        default=RequestsState.REQUESTED,
        verbose_name=_('Invitation__state'),
    )

    def __str__(self) -> str:  # pragma: no cover
        return f'{self.requester} -> {self.requested}'

    def __repr__(self) -> str:  # pragma: no cover
        return f'<{self.__class__.__name__}: {self.requester} -> {self.requested}>'

    @classmethod
    def type_is(cls, obj: object) -> TypeIs['Invitation']:
        """Check if the object is an Invitation.

        Returns:
            TypeIs[Invitation]: True if the object is an Invitation, False otherwise.
        """
        return isinstance(obj, cls)

    def clean(self) -> None:
        """Validation for the invitation class.

        Raises:
            InvitationAlreadyExists: Is raised when an invitation already exists between a requester and a requested entity.
            InvitationRejectedTimeout: Is raised when an invitation has been rejected and the timeout has not yet expired.
        """
        qs = Invitation.objects.filter(requester_id=self.requester_id, requested_id=self.requested_id).exclude(id=self.id)
        if qs.exists():
            if qs.filter(state=RequestsState.REQUESTED).exists():
                raise InvitationAlreadyExists(qs.first())
            last_rejection = (
                qs.filter(state=RequestsState.REJECTED).annotate(updated_last=Max('updated_at')).values('conference__timezone', 'updated_at').first()
            )
            if last_rejection:
                if last_rejection['updated_at'] > datetime.now(tz=last_rejection['conference__timezone']) - timedelta(
                    hours=settings.INVITATION_REJECTION_DEFAULT_TIMEOUT
                ):
                    raise InvitationRejectedTimeout
        super().clean()

    def accept(self, user: 'PlatformUser'):
        """Accept the invitation.

        Args:
            user (PlatformUser): The user who accepts the invitation.

        Raises:
            NotImplementedError: Raied when the invitation type is not yet implemented.
        """
        self.decision_by = user
        match self.type:
            case InvitationType.HABITAT:
                self.accept_habitat()
            case _:
                raise NotImplementedError

    def accept_habitat(self):
        """Accept the invitation for a joining a habitat.

        Raises:
            ValueError: I raised when the requested or requester type is not a habitat.
        """
        from core.models import Assembly  # pylint: disable=import-outside-toplevel

        if not Assembly.type_is(self.requested):  # pragma: no cover
            raise ValueError('Invalid requested type for invitation type')
        if not Assembly.type_is(self.requester):  # pragma: no cover
            raise ValueError('Invalid requested type for invitation type')
        if self.requester.hierarchy == Assembly.Hierarchy.REGULAR:
            habitat = self.requested
            assembly = self.requester
        else:
            habitat = self.requester
            assembly = self.requested
        parent_change = ActivityLogChange(old=assembly.parent.name, new=habitat.name) if assembly.parent else ActivityLogChange(new=habitat.name)
        old_children = ', '.join(habitat.children.all().values_list('name', flat=True))
        with transaction.atomic():
            assembly.parent = habitat
            assembly.save()

            self.state = RequestsState.ACCEPTED
            self.save()

            assembly.log_activity(
                self.decision_by,
                parent=parent_change,
            )
            habitat.log_activity(
                self.decision_by,
                children=ActivityLogChange(old=old_children, new=', '.join(habitat.children.all().values_list('name', flat=True))),
            )

    def reject(self, user: 'PlatformUser'):
        """Reject the invitation.

        Args:
            user (PlatformUser): The user rejecting the invitation.
        """
        self.decision_by = user
        self.state = RequestsState.REJECTED
        self.save()

    def withdraw(self, user: 'PlatformUser'):
        """Withdraw the invitation.

        Args:
            user (PlatformUser): The user withdrawing the invitation.
        """
        self.decision_by = user
        self.state = RequestsState.WITHDRAWN
        self.save()
