import contextlib

from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from django.core.validators import validate_slug
from django.db import models
from django.db.models import QuerySet
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _

from core.fields import ConferenceReference
from core.models.conference import Conference, ConferenceMember


class ConferenceTag(models.Model):
    class Meta:
        constraints = [
            models.UniqueConstraint(fields=['conference', 'slug'], name='unique_conference_tag'),
        ]
        verbose_name = _('ConferenceTag')
        verbose_name_plural = _('ConferenceTags')

    class Type(models.TextChoices):
        SIMPLE = 'simple', _('ConferenceTag__valuetype-simple')
        STRING = 'string', _('ConferenceTag__valuetype-string')
        INT = 'int', _('ConferenceTag__valuetype-int')
        BOOL = 'boolean', _('ConferenceTag__valuetype-bool')

    conference = ConferenceReference()
    slug = models.SlugField(help_text=_('ConferenceTag__slug__help'), verbose_name=_('ConferenceTag__slug'))
    is_public = models.BooleanField(default=False, help_text=_('ConferenceTag__is_public__help'), verbose_name=_('ConferenceTag__is_public'))

    value_type = models.CharField(
        max_length=20, choices=Type.choices, default=Type.SIMPLE, help_text=_('ConferenceTag__value_type__help'), verbose_name=_('ConferenceTag__value_type')
    )
    description = models.TextField(blank=True, null=True, help_text=_('ConferenceTag__description__help'), verbose_name=_('ConferenceTag__description'))

    def __str__(self):
        return self.slug


class TagItemManager(models.Manager):
    def conference_accessible(self, conference: 'Conference', member: ConferenceMember | None = None) -> QuerySet['TagItem']:
        return self.get_queryset().filter(tag__is_public=True).select_related('tag')


class TagItem(models.Model):
    class Meta:
        constraints = [models.UniqueConstraint(fields=['target_type', 'target_id', 'tag'], name='unique_target_tag')]

    tag = models.ForeignKey(ConferenceTag, related_name='+', on_delete=models.CASCADE)

    target_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    target_id = models.UUIDField()
    target = GenericForeignKey('target_type', 'target_id')

    _value_text = models.CharField(max_length=100, blank=True, null=True)
    _value_int = models.IntegerField(blank=True, null=True)

    objects = TagItemManager()

    @cached_property
    def slug(self):
        return self.tag.slug

    @property
    def value(self):
        if self.tag.value_type == ConferenceTag.Type.SIMPLE:
            return None

        if self.tag.value_type == ConferenceTag.Type.STRING:
            return self._value_text

        if self.tag.value_type == ConferenceTag.Type.INT:
            return self._value_int

        if self.tag.value_type == ConferenceTag.Type.BOOL:
            return self._value_int == 0

    @value.setter
    def value(self, new_value):
        self.set_value(new_value)

    def set_value(self, new_value):
        self._value_text = None
        self._value_int = None

        if self.tag.value_type == ConferenceTag.Type.SIMPLE:
            if new_value is not None:
                raise ValueError('Untyped tag must not be given a non-null value.')
            return

        if self.tag.value_type == ConferenceTag.Type.STRING:
            if not isinstance(new_value, str):
                raise ValueError('Tag expects a string value.')
            self._value_text = new_value
            return

        if self.tag.value_type == ConferenceTag.Type.INT:
            if not isinstance(new_value, int):
                raise ValueError('Tag expects a numeric value.')
            self._value_int = new_value
            return

        if self.tag.value_type == ConferenceTag.Type.BOOL:
            if isinstance(new_value, (bool, int)):
                self._value_int = 0 if new_value else 1
                return

            if isinstance(new_value, str):
                if new_value.lower() in ['y', 'j', 'yes', 'ja', 'true', 'on']:
                    self._value_int = 0
                    return

                if new_value.lower() in ['n', 'no', 'nein', 'false', 'off']:
                    self._value_int = 1
                    return

                raise ValueError(f'Cannot map string "{new_value}" to bool.')

            raise ValueError('Given value cannot be mapped to bool.')

        raise ValueError('Unknown tag value type. Cannot assign anything.')

    def clean_fields(self, **kwargs):
        super().clean_fields(**kwargs)

        if self.value is not None:
            if self.tag.value_type == ConferenceTag.Type.SIMPLE:
                raise ValidationError('Tag does not allow values.')

            elif self.tag.value_type == ConferenceTag.Type.INT:
                if not isinstance(self.value, int):
                    raise ValidationError("Tag's value must be integer.")

            elif self.tag.value_type == ConferenceTag.Type.BOOL:
                if self.value not in [True, False]:
                    raise ValidationError("Tag's value must be a boolean.")

    def __str__(self):
        if self.tag.value_type == ConferenceTag.Type.SIMPLE:
            return self.tag.slug

        return f'{self.tag.slug}={self.value}'


class TaggedItemMixin:
    @cached_property
    def sorted_tags(self):
        return sorted((t.slug for t in self.tags.all()), key=str.casefold)

    def has_tag(self, tag):
        if not isinstance(tag, ConferenceTag):
            tag = ConferenceTag.objects.get(conference=self.conference, slug=tag)

        return self.tags.filter(tag=tag).exists()

    def update_tags(self, new_tags: list[str], *, auto_create=False) -> tuple[set[str], set[str]]:
        new_set = set(new_tags)
        current_tags = set(self.tags.values_list('tag__slug', flat=True))
        removed_tags = current_tags.difference(new_set)
        added_tags = new_set.difference(current_tags)

        for added_tag in added_tags:
            self.add_tag(added_tag, auto_create=auto_create)
        for removed_tag in removed_tags:
            self.remove_tag(removed_tag)

        return (added_tags, removed_tags)

    def add_tag(self, tag, value=None, *, auto_create=False):
        if not isinstance(tag, ConferenceTag):
            try:
                tag = ConferenceTag.objects.get(conference=self.conference, slug=tag, is_public=True)
            except ConferenceTag.DoesNotExist as err:
                if auto_create:
                    tag = ConferenceTag(conference=self.conference, slug=tag, is_public=True, value_type=ConferenceTag.Type.SIMPLE)
                    tag.save()
                else:
                    raise err

        # TODO: check user's permission
        item, _created = self.tags.get_or_create(tag=tag)
        item.set_value(value)
        item.save()

    def remove_tag(self, tag, value=None):
        if not isinstance(tag, ConferenceTag):
            with contextlib.suppress(ConferenceTag.DoesNotExist):
                tag = ConferenceTag.objects.get(conference=self.conference, slug=tag, is_public=True)

        # TODO: check user's permission
        with contextlib.suppress(TagItem.DoesNotExist):
            self.tags.remove(self.tags.get(tag=tag))


def clean_tags(tags: str):
    # try to detect people delimiting tags with space instead of comma

    # allow empty tags
    if tags.strip() == '':
        return ''

    split_tags = tags.split(',')
    if tags is not None and len(split_tags) == 1 and len(tags.split(' ')) > 2:
        raise ValidationError(_('Tags__split_with_comma'))
    stripped_tags = [tag.strip() for tag in split_tags]
    for tag in stripped_tags:
        validate_slug(tag)

    return stripped_tags
