from django.core.exceptions import ValidationError
from django.test import TestCase

from core.choices import UserBadgeVisibility
from core.models.assemblies import Assembly
from core.models.badges import Badge, UserBadge
from core.models.conference import Conference, ConferenceMember
from core.models.users import PlatformUser


class BadgeTests(TestCase):
    def setUp(self):
        self.conference1 = Conference(slug='foo', name='Foo Conference')
        self.conference1.save()

        self.user = PlatformUser(username='bernd')
        self.user.save()
        self.assembly_user = PlatformUser(username='brot')
        self.assembly_user.save()
        ConferenceMember(conference=self.conference1, user=self.user).save()

        self.assembly1a = Assembly(conference=self.conference1, slug='fnord', name='Fnord Assembly', is_official=True)
        self.assembly1a.save()

    def test_creation(self):
        achievement = Badge(issuing_assembly=self.assembly1a, state=Badge.State.PUBLIC, name='Visited us.')
        achievement.save()

        easter_egg = Badge(conference=self.conference1, issuing_assembly=self.assembly1a, state=Badge.State.HIDDEN, name='Mischief Managed.')
        easter_egg.save()

        with self.assertRaises(ValidationError):
            other_conf = Conference(slug='other', name='Dummy')
            other_conf.save()
            should_fail = Badge(conference=other_conf, issuing_assembly=self.assembly1a, state=Badge.State.PUBLIC, name='Mischief Managed.')
            should_fail.save()

    def test_awarding(self):
        achievement = Badge(issuing_assembly=self.assembly1a, state=Badge.State.PUBLIC, name='Skills demonstrated')
        achievement.save()

        # award it to the user
        self.assertFalse(self.user.badges.exists())
        UserBadge.objects.redeem_badge(user=self.user, badge=achievement, issuer=self.assembly_user)
        self.assertEqual(1, self.user.badges.count())
        ub = self.user.badges.get(badge=achievement)
        self.assertEqual(UserBadgeVisibility.PRIVATE, ub.visibility)
        # TODO: check that notification has been sent

        # user makes badge visible
        ub.visibility = UserBadgeVisibility.PUBLIC
        ub.save()

        # award it a second time, should remain visible
        UserBadge.objects.redeem_badge(user=self.user, badge=achievement, issuer=self.assembly_user)
        self.assertEqual(1, self.user.badges.count())
        ub.refresh_from_db()
        self.assertEqual(UserBadgeVisibility.PUBLIC, ub.visibility)
        # TODO: check that no notification has been sent
