from datetime import UTC, datetime, timedelta

from django.test import TestCase, override_settings
from django.utils.timezone import now

from core.models.assemblies import Assembly
from core.models.conference import Conference, ConferenceExportCache


class ScheduleTests(TestCase):
    def setUp(self):
        self.conference = Conference(
            slug='foo',
            name='Foo Conference',
            start=datetime(2021, 12, 1, 9, 0, 0, tzinfo=UTC),
            end=datetime(2021, 12, 31, 15, 00, 00, tzinfo=UTC),
        )
        self.conference.save()

        self.assembly = Assembly(conference=self.conference, slug='fnord', name='Fnord Assembly', is_official=True)
        self.assembly.save()

    @override_settings(SCHEDULES_SUPPORT_FILE_PROTOCOL=True)
    def test_schedule_cache(self):
        # there should be no cached items at all
        self.assertFalse(ConferenceExportCache.objects.filter(type=ConferenceExportCache.Type.SCHEDULE).exists())

        # create an event
        e = self.conference.events.create(assembly=self.assembly, name='Fnord', is_public=True)
        e.save()

        # creating the event should not have created a cache entry
        self.assertFalse(ConferenceExportCache.objects.filter(type=ConferenceExportCache.Type.SCHEDULE).exists())

        # create one manually
        c = ConferenceExportCache(conference=self.conference, type=ConferenceExportCache.Type.SCHEDULE, tag='')
        c.save()

        # check default data
        self.assertTrue(c.needs_regeneration)
        self.assertTrue(c.is_dirty)

        # simulate successful generation
        c.needs_regeneration = False
        c.data = 'FNORD'
        c.last_generated = now()
        c.save()

        # this should have reset the 'is_dirty' flag
        self.assertFalse(c.is_dirty)

        # update the event
        e.last_update = c.last_generated + timedelta(seconds=1)  # ensure event's last_update is newer
        e.save()

        # now the flag should react to the event updated after 'last_generation'
        self.assertTrue(c.is_dirty)
