import re
import uuid
from urllib.parse import quote

from django.conf import settings
from django.test import TestCase
from django.test.utils import override_settings
from django.urls import reverse

from core.markdown import render_markdown, render_markdown_ex
from core.models import Conference, ConferenceTag, PlatformUser, StaticPage

TEST_CONF_ID = uuid.uuid4()


@override_settings(ALLOWED_HOSTS=['hub.test'], SELECTED_CONFERENCE_ID=TEST_CONF_ID)
class MarkdownTest(TestCase):
    def test_url_class(self):
        conf = Conference(name='foo', id=TEST_CONF_ID)
        conf.save()

        tests = [
            ('https://localhost/', False),
            ('https://localhost/foo', False),
            ('https://localhost/foo/', False),
            ('https://localhost/foo.bar', False),
            ('https://some.test/', False),
            ('https://some.test/foo', False),
            ('https://some.test/foo.bar', False),
            ('/foo/bar', True),
            ('foo/bar', True),
            ('asdf', True),
            ('https://hub.test/', True),
            ('https://hub.test/foo', True),
            ('https://hub.test/foo.bar', True),
        ]
        for test_url, should_be_local in tests:
            with self.subTest(test_url):
                if should_be_local:
                    class_ = 'internal'
                    href = test_url
                else:
                    href = settings.PLAINUI_DEREFERER_URL.format(quoted_target=quote(test_url))
                    class_ = 'external'

                self.assertEqual(render_markdown(conf, f'[.]({test_url})'), f'<p><a href="{href}" class="{class_}">.</a></p>\n')
                self.assertEqual(render_markdown(conf, f'[.][1]\n\n[1]: {test_url}'), f'<p><a href="{href}" class="{class_}">.</a></p>\n')

        forbidden = [
            'data:text/html;charset=utf-8;base64,PGh0bWw+PGhlYWQ+PHRpdGxlPnRlc3Q8L3RpdGxlPjwvaGVhZD48Ym9keT48aDE+VGVzdDwvaDE+PC9ib2R5PjwvaHRtbD4=',
            'javascript:alert(1)',
        ]
        for test_url in forbidden:
            with self.subTest(test_url):
                self.assertEqual(render_markdown(conf, f'[.]({test_url})'), '<p><a href="" class="internal">.</a></p>\n')
                self.assertEqual(render_markdown(conf, f'[.][1]\n\n[1]: {test_url}'), '<p><a href="" class="internal">.</a></p>\n')

        self.assertEqual(render_markdown_ex(conf, '[.](https://hub.test/foo)').linked_urls, {'https://hub.test/foo'})

    @override_settings(SELECTED_CONFERENCE_ID=TEST_CONF_ID)
    def test_page_links(self):
        conf = Conference(name='foo', id=TEST_CONF_ID)
        conf.save()
        existing_page = StaticPage(conference=conf, slug='existing_page', title='Existing Page')
        existing_page.save()

        self.assertEqual(render_markdown(conf, '[[foo]]'), '<p><a href="/en/wiki/foo/edit" class="internal-nonexist">foo</a></p>\n')
        self.assertEqual(render_markdown(conf, '[[foo | bar]]'), '<p><a href="/en/wiki/foo/edit" class="internal-nonexist">bar</a></p>\n')
        self.assertEqual(render_markdown(conf, '[[foo | *bar*]]'), '<p><a href="/en/wiki/foo/edit" class="internal-nonexist"><em>bar</em></a></p>\n')

        self.assertEqual(render_markdown(conf, '[[existing_page]]'), '<p><a href="/en/wiki/existing_page/" class="internal">Existing Page</a></p>\n')
        self.assertEqual(render_markdown(conf, '[[existing_page |]]'), '<p><a href="/en/wiki/existing_page/" class="internal">Existing Page</a></p>\n')
        self.assertEqual(render_markdown(conf, '[[existing_page| ]]'), '<p><a href="/en/wiki/existing_page/" class="internal">Existing Page</a></p>\n')
        self.assertEqual(render_markdown(conf, '[[existing_page | ]]'), '<p><a href="/en/wiki/existing_page/" class="internal">Existing Page</a></p>\n')
        self.assertEqual(render_markdown(conf, '[[existing_page | bar]]'), '<p><a href="/en/wiki/existing_page/" class="internal">bar</a></p>\n')
        self.assertEqual(render_markdown(conf, '[[existing_page | *bar*]]'), '<p><a href="/en/wiki/existing_page/" class="internal"><em>bar</em></a></p>\n')

        self.assertEqual(render_markdown_ex(conf, '[[foo]]').linked_page_slugs, {'foo'})
        self.assertEqual(render_markdown_ex(conf, '[[existing_page]]').linked_page_slugs, {'existing_page'})

    @override_settings(SELECTED_CONFERENCE_ID=TEST_CONF_ID)
    def test_profile_link(self):
        conf = Conference(name='foo', id=TEST_CONF_ID)
        conf.save()
        existing_user = PlatformUser(slug='existing_user')
        existing_user.save()

        self.assertEqual(render_markdown(conf, '@existing_user'), '<p><a href="/en/user/existing_user/" class="internal">@existing_user</a></p>\n')
        self.assertEqual(render_markdown(conf, '@nonexisting_user'), '<p><a href="/en/user/nonexisting_user/" class="internal">@nonexisting_user</a></p>\n')
        self.assertEqual(render_markdown(conf, 'foo@bar'), '<p>foo@bar</p>\n')
        self.assertEqual(render_markdown(conf, '@"'), '<p>@"</p>\n')

        self.assertEqual(render_markdown_ex(conf, '@existing_user').linked_user_slugs, {'existing_user'})
        self.assertEqual(render_markdown_ex(conf, '@nonexisting_user').linked_user_slugs, {'nonexisting_user'})

    @override_settings(SELECTED_CONFERENCE_ID=TEST_CONF_ID)
    def test_tag_link(self):
        conf = Conference(name='foo', id=TEST_CONF_ID)
        conf.save()
        existing_tag = ConferenceTag(conference=conf, slug='existing_tag')
        existing_tag.save()

        self.assertEqual(render_markdown(conf, '#existing_tag'), '<p><a href="/en/tag/existing_tag/" class="internal">#existing_tag</a></p>\n')
        self.assertEqual(render_markdown(conf, '#nonexisting_tag'), '<p><a href="/en/tag/nonexisting_tag/" class="internal">#nonexisting_tag</a></p>\n')
        self.assertEqual(render_markdown(conf, '#"'), '<p>#"</p>\n')
        self.assertEqual(render_markdown(conf, 'foo#bar'), '<p>foo#bar</p>\n')

        self.assertEqual(render_markdown_ex(conf, '#existing_tag').applied_tags, {'existing_tag'})
        self.assertEqual(render_markdown_ex(conf, '#nonexisting_tag').applied_tags, {'nonexisting_tag'})

    @override_settings(SELECTED_CONFERENCE_ID=TEST_CONF_ID)
    def test_markdownextensions(self):
        conf = Conference(name='foo', id=TEST_CONF_ID)
        conf.save()

        md_markup = (
            'Dies ist ein Test für @username Highlights mit einem @ vor dem Namen.\n'
            '@highlight sollte auch am Zeilenanfang sowie am Zeilenende funktionieren @really\n'
            'Insbesondere sollte er aber nicht auf demo@emailadres.se matchen, auch nicht'
            'in einem [link](https://foo:bar@domain.tld/).'
        )
        md_rendered = render_markdown(conf, md_markup)

        # check for the correct links generated
        re_userlink = re.compile(reverse('plainui:user', kwargs={'user_slug': 'unittest'}).replace('unittest', r'(?P<uslug>[\w\d]+)'))
        parsed_links_slugs = [m.group('uslug') for m in re_userlink.finditer(md_rendered)]

        # verify expected results
        self.assertListEqual(['username', 'highlight', 'really'], parsed_links_slugs)

        # triple sanity check: we expected four links total (one of them not to a user)
        self.assertEqual(4, len(re.findall('<a ', md_rendered)))
