from django.contrib.auth.models import AnonymousUser
from django.contrib.contenttypes.models import ContentType
from django.test import TestCase

from core.models.assemblies import Assembly
from core.models.conference import Conference
from core.models.events import Event
from core.models.tags import ConferenceTag, TaggedItemMixin, TagItem


class TaggingTests(TestCase):
    fixtures = ['anhalter.json']

    def setUp(self):
        self.conference = Conference.objects.get(slug='vogc')
        self.user = AnonymousUser()

    def _testTagManagement(self, target):
        self.assertIsInstance(target, TaggedItemMixin)

        foo_tag = ConferenceTag.objects.get(conference=self.conference, slug='foo')
        bar_tag = ConferenceTag.objects.get(conference=self.conference, slug='bar')
        stars_tag = ConferenceTag.objects.get(conference=self.conference, slug='michelinstars')

        # we start without any tags
        self.assertEqual(target.tags.count(), 0)

        # has_tag() doesn't see any tags, too
        self.assertFalse(target.has_tag(foo_tag))
        self.assertFalse(target.has_tag('bar'))

        # add a tag and check has_tag() and .tags
        target.add_tag('foo')
        self.assertTrue(target.has_tag('foo'))
        self.assertTrue(target.has_tag(foo_tag))
        self.assertTrue(any(tag.slug == 'foo' for tag in target.tags.all()))

        # add the same tag again (shouldn't throw an error)
        target.add_tag('foo')
        target.add_tag(foo_tag)

        # remove the tag and check again
        target.remove_tag(foo_tag)
        self.assertFalse(target.has_tag('foo'))
        self.assertFalse(any(tag.slug == 'foo' for tag in target.tags.all()))

        # remove an already removed tag twice (shouldn't throw an error)
        target.remove_tag(foo_tag)

        # add two tags
        self.assertFalse(target.has_tag('bar'))
        target.add_tag(bar_tag)
        self.assertTrue(target.has_tag('bar'))
        target.add_tag(foo_tag)
        self.assertEqual(target.tags.count(), 2)
        self.assertTrue(all(tag.value is None for tag in target.tags.all()))

        # add value-type tags
        target.add_tag(stars_tag.slug, 42)
        self.assertTrue(target.has_tag(stars_tag))
        self.assertTrue(any(tag.slug == stars_tag.slug and tag.value == 42 for tag in target.tags.all()))
        target.remove_tag(stars_tag)

        # add value-type tags without value
        with self.assertRaises(ValueError):
            target.add_tag(stars_tag)

        # add value-type tags with incompatible value type
        with self.assertRaises(ValueError):
            target.add_tag(stars_tag, 'foo')

        # add non-value-type tag with value
        with self.assertRaises(ValueError):
            target.add_tag(bar_tag, 23)

        # auto-create non-existant tag
        self.assertFalse(ConferenceTag.objects.filter(conference=self.conference, slug='fnord').exists())
        target.add_tag('fnord', auto_create=True)
        self.assertTrue(ConferenceTag.objects.filter(conference=self.conference, slug='fnord').exists())
        self.assertTrue(target.has_tag('fnord'))
        target.remove_tag('fnord')
        self.assertTrue(ConferenceTag.objects.filter(conference=self.conference, slug='fnord').exists())
        self.assertFalse(target.has_tag('fnord'))

        # TODO: tag.is_public

    def testEventTagging(self):
        event = Event.objects.filter(conference=self.conference).first()
        self._testTagManagement(event)

    def testAssemblyTagging(self):
        assembly = Assembly.objects.filter(conference=self.conference).first()
        self._testTagManagement(assembly)


class TagItemTests(TestCase):
    fixtures = ['anhalter.json']

    def setUp(self):
        self.conference = Conference.objects.get(slug='vogc')
        self.user = AnonymousUser()
        self.event = Event.objects.filter(conference=self.conference).first()
        self.tag_items = self._createTagItems()

    def _createTagItem(self, tag_slug: str) -> TagItem:
        tag = ConferenceTag.objects.get(conference=self.conference, slug=tag_slug)
        return TagItem(tag=tag, target_type=ContentType.objects.get_for_model(type(self.event)), target_id=self.event.id)

    def _createTagItems(self):
        result = {}
        for tag_slug in ['foo', 'bar', 'michelinstars', 'guter_geschmack', 'autor_des_besten_gedichts']:
            result[tag_slug] = self._createTagItem(tag_slug)

        return result

    def _validate_value_type(self, tag_item: TagItem):
        value_type = tag_item.tag.value_type
        value = tag_item.value
        if value is None:
            return
        if value_type == ConferenceTag.Type.SIMPLE:
            self.fail('Simple tag item did not return None as value')
        if value_type == ConferenceTag.Type.STRING:
            self.assertTrue(isinstance(value, str), 'String tag item did not return None or a string as value')
        if value_type == ConferenceTag.Type.INT:
            self.assertTrue(isinstance(value, int), 'Int tag item did not return None or an int as value')
        if value_type == ConferenceTag.Type.BOOL:
            self.assertTrue(isinstance(value, bool), 'Bool tag item did not return None or a bool as value')

    def test_get_value_check_type_uninitialized(self):
        for tag_item in self.tag_items.values():
            self._validate_value_type(tag_item)

    def test_get_value_check_type_after_set(self):
        for tag_item in self.tag_items.values():
            value_type = tag_item.tag.value_type
            if value_type == ConferenceTag.Type.STRING:
                tag_item.value = 'Test'
            if value_type == ConferenceTag.Type.INT:
                tag_item.value = 42
            if value_type == ConferenceTag.Type.BOOL:
                tag_item.value = True
            self._validate_value_type(tag_item)

    def test_value_check_get_after_set_intended(self):
        for tag_item in self.tag_items.values():
            value_type = tag_item.tag.value_type
            value = None
            if value_type == ConferenceTag.Type.STRING:
                value = 'Test'
            if value_type == ConferenceTag.Type.INT:
                value = 42
            if value_type == ConferenceTag.Type.BOOL:
                value = True
            tag_item.value = value
            self.assertEqual(tag_item.value, value)

    def test_value_check_boolean_coerce(self):
        tag = self.tag_items['guter_geschmack']
        for v in [False, 0, 'n', 'no', 'nein', 'false', 'off']:
            tag.value = True
            self.assertTrue(tag.value)
            tag.value = v
            self.assertFalse(tag.value, f'Value {v} did not yield false')

        for v in [True, -2, -1, 1, 2, 3, 'y', 'j', 'yes', 'ja', 'true', 'on']:
            tag.value = False
            self.assertFalse(tag.value)
            tag.value = v
            self.assertTrue(tag.value, f'Value {v} did not yield true')

    def test_value_set_check_simple_with_wrong_argument(self):
        tag = self.tag_items['foo']
        with self.assertRaises(ValueError):
            tag.value = self.user
        with self.assertRaises(ValueError):
            tag.value = 'Test'
        with self.assertRaises(ValueError):
            tag.value = 42
        with self.assertRaises(ValueError):
            tag.value = True

    def test_value_set_check_string_with_wrong_argument(self):
        tag = self.tag_items['autor_des_besten_gedichts']
        with self.assertRaises(ValueError):
            tag.value = None
        with self.assertRaises(ValueError):
            tag.value = self.user
        with self.assertRaises(ValueError):
            tag.value = 42
        with self.assertRaises(ValueError):
            tag.value = True

    def test_value_set_check_int_with_wrong_argument(self):
        tag = self.tag_items['michelinstars']
        with self.assertRaises(ValueError):
            tag.value = None
        with self.assertRaises(ValueError):
            tag.value = self.user
        with self.assertRaises(ValueError):
            tag.value = 'Test'
        # Python coerces bool to int by itself

    def test_value_set_check_bool_with_wrong_argument(self):
        tag = self.tag_items['guter_geschmack']
        with self.assertRaises(ValueError):
            tag.value = None
        with self.assertRaises(ValueError):
            tag.value = self.user
        with self.assertRaises(ValueError):
            tag.value = 'Test'
        with self.assertRaises(ValueError):
            tag.value = 'foo'
        with self.assertRaises(ValueError):
            tag.value = ''
        with self.assertRaises(ValueError):
            tag.value = 'bar'
