from datetime import UTC, datetime, timedelta

import jwt

from django.test import TestCase, override_settings

from core.models.conference import Conference
from core.models.ticket import ConferenceMemberTicket, TicketValidationError
from core.models.users import PlatformUser

_GOOD_SECRET = 'F00Preti%'
_BAD_SECRET = 'DämlicherFlanders'


@override_settings(PRETIX_SECRET_KEY=_GOOD_SECRET)
class PretixTests(TestCase):
    fixtures = ['anhalter.json']

    def setUp(self):
        self.conference = Conference.objects.get(slug='vogc')
        self.foo_user = PlatformUser(username='foo')
        self.foo_user.save()
        self.pretix_jwt_basepayload = {
            'aud': 'vogc',
            'exp': datetime.now(UTC) + timedelta(minutes=1),
            'iat': datetime.now(UTC) - timedelta(hours=1),
            'iss': 'tickets.events.ccc.de',
        }

    def test_good(self):
        token = jwt.encode({**self.pretix_jwt_basepayload, 'uid': 'asdf1234'}, _GOOD_SECRET)

        # using the token for the first time should succeed
        ticket = ConferenceMemberTicket.redeem_pretix_ticket(conference=self.conference, user=self.foo_user, token=token)
        self.assertNotEqual(None, ticket)
        self.assertEqual(ticket.ident, 'asdf1234')
        self.assertEqual(ticket.conference_id, self.conference.id)
        self.assertEqual(ticket.user_id, self.foo_user.id)

        # trying the same token again must fail
        with self.assertRaises(TicketValidationError), self.assertLogs('core.models.ticket', 'WARNING'):
            ConferenceMemberTicket.redeem_pretix_ticket(conference=self.conference, user=self.foo_user, token=token)

    def test_two_per_user(self):
        token1 = jwt.encode({**self.pretix_jwt_basepayload, 'uid': 'asdf1234'}, _GOOD_SECRET)
        ConferenceMemberTicket.redeem_pretix_ticket(conference=self.conference, user=self.foo_user, token=token1)

        # validating a second token must fail (even if the token itself is valid)
        token2 = jwt.encode({**self.pretix_jwt_basepayload, 'uid': 'jklö5678'}, _GOOD_SECRET)
        with self.assertRaises(TicketValidationError), self.assertLogs('core.models.ticket', 'WARNING'):
            ConferenceMemberTicket.redeem_pretix_ticket(conference=self.conference, user=self.foo_user, token=token2)

        # verify that the second token works on another user
        bar_user = PlatformUser(username='bar')
        bar_user.save()
        ConferenceMemberTicket.redeem_pretix_ticket(conference=self.conference, user=bar_user, token=token2)

    def test_invalid_signature(self):
        token = jwt.encode({**self.pretix_jwt_basepayload, 'uid': 'asdf1234'}, _BAD_SECRET)

        with self.assertRaises(TicketValidationError), self.assertLogs('core.models.ticket', 'WARNING'):
            ConferenceMemberTicket.redeem_pretix_ticket(conference=self.conference, user=self.foo_user, token=token)

    def test_invalid_fields(self):
        # misspell 'uid'
        token = jwt.encode({**self.pretix_jwt_basepayload, 'userid': 'asdf1234'}, _GOOD_SECRET)
        with self.assertRaises(TicketValidationError), self.assertLogs('core.models.ticket', 'WARNING'):
            ConferenceMemberTicket.redeem_pretix_ticket(conference=self.conference, user=self.foo_user, token=token)

        # unmatching 'aud' (audience, conference slug)
        token = jwt.encode({**self.pretix_jwt_basepayload, 'aud': 'FNORD', 'userid': 'asdf1234'}, _GOOD_SECRET)
        with self.assertRaises(TicketValidationError), self.assertLogs('core.models.ticket', 'WARNING'):
            ConferenceMemberTicket.redeem_pretix_ticket(conference=self.conference, user=self.foo_user, token=token)
