from django.core.exceptions import ValidationError
from django.test import TestCase
from django.urls import reverse

from core.models.conference import Conference, ConferenceMember
from core.models.users import PlatformUser, UserCommunicationChannel


class UserCommunicationChannelStaticTests(TestCase):
    def check_good_bad(self, what, func, good, bad):
        for g in good:
            try:
                self.assertIsNone(func(g), msg=f'{g} should be accepted as {what}')
            except Exception as err:
                self.fail(f'{g} should be accepted as {what} but raised an exception: {err}')

        for b in bad:
            with self.assertRaises(ValidationError, msg=f"{b} shouldn't be accepted as {what}"):
                func(b)

    def test_mail_validation(self):
        good_addresses = [  # most found on https://matrix.org/docs/spec/
            'foo@example.org',
            'foo.bar@example.org',
            'foo.bar+hint@example.org',
        ]
        bad_addresses = [
            '',
            'foo.bar at example.org',
            'foo.bar at example dot org',
        ]

        self.check_good_bad('email address', UserCommunicationChannel.validate_email, good_addresses, bad_addresses)

    def test_matrixroom_validation(self):
        good_uris = [  # most found on https://matrix.org/docs/spec/
            '!qporfwt:matrix.org',
            '#matrix:example.org',
            '!4rguxf:matrix.org',
            'https://matrix.to/#/%23somewhere%3Aexample.org',
            'https://matrix.to/#/!somewhere%3Aexample.org',
        ]
        bad_uris = [
            '',
            'matrix.org/example',
            '#matrix on example.org',
            '#matrix@example.org',
            'https://example.org/matrix',
        ]

        self.check_good_bad('Matrix room URI', UserCommunicationChannel.validate_matrix_room_url, good_uris, bad_uris)

    def test_phonenumber_validation(self):
        good_numbers = [
            '+491234567890',
            '+49 1234 567890',
            '+49 1234 56789-0',
            '+1-234-567-890',
            '001-234-567-890',
            '0049 700 VANITY',
            '+49-800-VANITY',
        ]
        bad_numbers = [
            '',
            'foo',
            '12345',
            '++1-234-567-890',
            '01-234-567-890',
            '01234 567890',
            '+4989VANITY',
        ]

        self.check_good_bad('phone number', UserCommunicationChannel.validate_phone_number, good_numbers, bad_numbers)

    def test_activitypub_validation(self):
        good_addresses = [
            'https://chaos.social/@ordnung',
            'https://chaos.social/@events/113499399916580165',
            'https://social.mrtoto.net/@apptest',
            'https://social.mrtoto.net/@apptest/113085273939532302',
        ]
        bad_addresses = [
            '',
            'foo',
            'chaos.social/ordnung',
            'chaos.social/@ordnung',
            'ftp://ccc.de',
            'hub@cccv.de',
        ]
        self.check_good_bad('ActivityPub URL', UserCommunicationChannel.validate_activitypub_url, good_addresses, bad_addresses)


class UserCommunicationChannelTests(TestCase):
    def setUp(self):
        self.conference = Conference(slug='test', name='Test Conference')
        self.conference.save()
        self.user = PlatformUser(username='bernd')
        self.user.save()
        ConferenceMember(conference=self.conference, user=self.user).save()

        self.mail_channel = UserCommunicationChannel(user=self.user, channel=UserCommunicationChannel.Channel.MAIL, address='foo.bar@example.com')
        self.dect_channel = UserCommunicationChannel(user=self.user, channel=UserCommunicationChannel.Channel.DECT, address='4711')

    def test_clean(self):
        with self.assertRaises(Exception):
            c = UserCommunicationChannel(user=self.user, channel=UserCommunicationChannel.Channel.MATRIX)
            c.clean()

        self.mail_channel.clean()
        self.dect_channel.clean()

    def test_can_notify(self):
        self.assertTrue(self.mail_channel.can_notify)
        self.assertFalse(self.dect_channel.can_notify)

        with self.assertRaises(Exception):
            self.dect_channel.send('foo', 'bar')


class UserRegistrationTests(TestCase):
    username = 'TestUser'
    password = 'UanaNeje10n!@#'
    email = 'a@example.com'

    def setUp(self) -> None:
        self.conference = Conference(slug='test', name='Test Conference')
        self.conference.save()

    def test_username_not_case_sensitive(self):
        registration_url = reverse('plainui:signup')
        response = self.client.post(
            registration_url,
            {
                'username': self.username,
                'email': self.email,
                'password1': self.password,
                'password2': self.password,
            },
        )
        self.assertEqual(response.status_code, 302)
        self.assertTrue(PlatformUser.objects.filter(username='TestUser').exists())
        self.user = PlatformUser.objects.get(username__iexact='TestUser')
        self.user.is_active = True
        self.user.save()
        # Test with matching case username
        self.assertTrue(self.client.login(username='TestUser', password=self.password))
        self.client.logout()
        # Test with lowercase
        self.assertTrue(self.client.login(username='testuser', password=self.password))

    def test_username_non_slug_characters_only(self):
        username = '@'
        registration_url = reverse('plainui:signup')
        response = self.client.post(
            registration_url,
            {
                'username': username,
                'email': self.email,
                'password1': self.password,
                'password2': self.password,
            },
        )
        self.assertEqual(response.status_code, 302)
        user = PlatformUser.objects.get(username=username)
        self.assertTrue(len(user.slug) > 0)
