from unittest.mock import Mock

from django.core.exceptions import ValidationError
from django.test import TestCase

from core.validators import FileSizeValidator, ImageDimensionValidator


class ValidateFileSizeTests(TestCase):
    def test_does_not_raise_when_size_meets_requirements(self):
        mock_file = Mock(size=100)
        validator = FileSizeValidator(2)

        self.assertEqual(mock_file, validator(mock_file))

    def test_raises_if_file_size_is_too_large(self):
        mock_file = Mock(size=100 * 1024 * 1024)
        validator = FileSizeValidator(2)

        with self.assertRaises(ValidationError):
            self.assertEqual(mock_file, validator(mock_file))


class ValidateImageSizeTests(TestCase):
    def test_does_not_raise_when_image_meets_requirements(self):
        mock_image = Mock(width=10, height=10)
        validator = ImageDimensionValidator(min_size=(10, 10), max_size=(10, 10), square=True)

        self.assertEqual(mock_image, validator(mock_image))

    def test_raises_if_image_is_too_small_width(self):
        mock_image = Mock(width=9, height=9)
        validator = ImageDimensionValidator(
            min_size=(10, None),
        )

        with self.assertRaises(ValidationError) as context:
            validator(image=mock_image)

        self.assertEqual(len(context.exception.error_list), 2)

    def test_raises_if_image_is_too_small_height(self):
        mock_image = Mock(width=9, height=9)
        validator = ImageDimensionValidator(
            min_size=(None, 10),
        )

        with self.assertRaises(ValidationError) as context:
            validator(image=mock_image)

        self.assertEqual(len(context.exception.error_list), 2)

    def test_raises_if_image_is_too_small_both(self):
        mock_image = Mock(width=9, height=9)
        validator = ImageDimensionValidator(
            min_size=(10, 10),
        )

        with self.assertRaises(ValidationError) as context:
            validator(image=mock_image)

        self.assertEqual(len(context.exception.error_list), 3)

    def test_raises_if_image_is_too_large_width(self):
        mock_image = Mock(width=11, height=11)
        validator = ImageDimensionValidator(max_size=(10, None))

        with self.assertRaises(ValidationError) as context:
            validator(image=mock_image)

        self.assertEqual(len(context.exception.error_list), 2)

    def test_raises_if_image_is_too_large_height(self):
        mock_image = Mock(width=11, height=11)
        validator = ImageDimensionValidator(max_size=(None, 10))

        with self.assertRaises(ValidationError) as context:
            validator(image=mock_image)

        self.assertEqual(len(context.exception.error_list), 2)

    def test_raises_if_image_is_too_large_both(self):
        mock_image = Mock(width=11, height=11)
        validator = ImageDimensionValidator(max_size=(10, 10))

        with self.assertRaises(ValidationError) as context:
            validator(image=mock_image)

        self.assertEqual(len(context.exception.error_list), 3)

    def test_raises_if_image_is_not_square(self):
        mock_image = Mock(width=11, height=12)
        validator = ImageDimensionValidator(square=True)

        with self.assertRaises(ValidationError) as context:
            validator(image=mock_image)

        self.assertEqual(len(context.exception.error_list), 1)
