import re
from datetime import datetime, timedelta
from typing import Any

from jinja2 import Environment, pass_context
from jinja2.runtime import Context

from django.conf import settings
from django.contrib.humanize.templatetags.humanize import NaturalTimeFormatter
from django.contrib.messages import get_messages
from django.http import HttpRequest
from django.templatetags.static import static
from django.urls import reverse
from django.utils import timezone
from django.utils.formats import localize
from django.utils.functional import LazyObject
from django.utils.html import json_script
from django.utils.http import urlencode
from django.utils.timezone import localdate, localtime
from django.utils.translation import get_language, gettext, ngettext
from modeltranslation.fields import build_localized_fieldname
from modeltranslation.settings import AVAILABLE_LANGUAGES

from core.models import PlatformUser, UserBadge
from core.templatetags.hub_absolute import hub_absolute, hub_absolute_self
from core.templatetags.update_parameters import update_parameters

from plainui.utils import fetch_wiki_page


def url(name, *args, current_app=None, **kwargs):
    return reverse(name, args=args, kwargs=kwargs, current_app=current_app)


def browser_type(request):
    user_agent = request.META.get('HTTP_USER_AGENT', '')
    if 'MSIE' in user_agent or 'Trident/7.0;' in user_agent:
        return 'ie'
    if 'CERN-NextStep-WorldWideWeb' in user_agent:
        return 'historic'
    return user_agent


def num_of_notifications(request):
    return num_of_unread_messages(request) + num_of_pending_badges(request)


def lighten(hex_colour, factor):
    if not isinstance(hex_colour, str) or not re.match('#[0-9a-f]{6}', hex_colour, re.IGNORECASE):
        # fall back to black in case of an invalid colour
        return '#000000'

    colour = hex_colour[1:]

    hex_red = int(255 - (255 - int(colour[0:2], base=16)) * (1 - factor))
    hex_green = int(255 - (255 - int(colour[2:4], base=16)) * (1 - factor))
    hex_blue = int(255 - (255 - int(colour[4:6], base=16)) * (1 - factor))

    return f'#{hex_red:02x}{hex_green:02x}{hex_blue:02x}'


def calculate_luminance(hex_colour):
    if not isinstance(hex_colour, str) or not re.match('#[0-9a-f]{6}', hex_colour, re.IGNORECASE):
        # fall back to 0 in case of an invalid colour
        return 0

    colour = hex_colour[1:]

    hex_red = int(colour[0:2], base=16)
    hex_green = int(colour[2:4], base=16)
    hex_blue = int(colour[4:6], base=16)

    return hex_red * 0.2126 + hex_green * 0.7152 + hex_blue * 0.0722


def calculate_text_colour(hex_colour):
    luminance = calculate_luminance(hex_colour)

    if luminance < 140:
        return '#ffffff'

    return '#000000'


def num_of_unread_messages(request):
    user = request.user
    if not user.is_authenticated:
        return 0
    return user.received_messages.filter(was_read=False).count()


def num_of_pending_badges(request):
    user = request.user
    if not user.is_authenticated:
        return 0
    return len(UserBadge.objects.filter(user=user, accepted_by_user=False).select_related('badge'))


def custom_timedelta(tdelta):
    return NaturalTimeFormatter.string_for(tdelta)


def custom_timedelta_short(tdelta: timedelta):
    if not isinstance(tdelta, timedelta):
        return ''
    s = tdelta.seconds
    h = s // 3600
    s -= h * 3600
    m = s // 60
    s -= m * 60
    return f'{h:02d}:{m:02d}:{s:02d}'


def custom_strftime(date):
    if not isinstance(date, datetime):
        return ''

    return localize(localtime(date))


def custom_strftimehm(date):
    if not isinstance(date, datetime):
        return ''

    return localtime(date).strftime('%H:%M')


def custom_strfdate(date):
    if not isinstance(date, datetime):
        return ''

    return localize(localdate(date))


def custom_strfdates(date):
    if not isinstance(date, datetime):
        return ''

    if get_language() == 'de':
        return localtime(date).strftime('%d.%m.%Y')

    return localdate(date).strftime('%x')


def custom_weekday_abbrev(date):
    if not isinstance(date, datetime):
        return ''

    return date.strftime('%a')


link_icon_map = {
    'website': 'globe',
    'chat': 'chat-fill',
    'bbb': 'headset',
    'jitsi': 'headset',
    'pad': 'pencil-fill',
    'media.ccc.de': 'play-circle-fill',
    'video': 'camera-reels-fill',
    'audio': 'volume-up-fill',
}


def link_icon(link):
    if not link:
        return 'link'

    return link_icon_map.get(link.link_type, 'link')


# set up an internal represenative for an unset variable as parameter for show_vars()
# which cannot use a default of None as that is a valid value for the variable
_UNSET = object()


@pass_context
def show_vars(ctx, var=_UNSET):
    if var is _UNSET:
        var = ctx

    if isinstance(var, LazyObject):
        var = var._wrapped

    try:
        if not callable(getattr(var, 'items', None)):
            var = var.__dict__

        ret = ''
        for k, v in var.items():
            ret += f'{k!r}: {v!r}\n'
        return ret

    except (AttributeError, TypeError):
        return repr(var)


@pass_context
def css_scope(ctx):
    request = ctx['request']

    return '38c3'

    if (request.user.is_authenticated and request.user.high_contrast) or request.session.get('high_contrast', False):
        return 'hub-high-contrast'

    return 'hub'


@pass_context
def active_theme(ctx):
    request = ctx['request']
    if 'theme_override' in ctx:
        return ctx['theme_override']
    theme = request.session.get('theme', PlatformUser.Theme.DARK)
    return theme


@pass_context
def unique_id(ctx):
    return ctx.get_next_id()


@pass_context
def _fetch_wiki_page(ctx, *args, **kwargs):
    conf = kwargs.pop('conference', ctx['conf'])
    return fetch_wiki_page(conf, *args, **kwargs)


def translated_fields_for_field(field_name):
    return [build_localized_fieldname(field_name, lang) for lang in AVAILABLE_LANGUAGES]


def field_translation_languages():
    return AVAILABLE_LANGUAGES


class MyContext(Context):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.next_id = 0

    def get_next_id(self):
        res = 'id_' + str(self.next_id)
        self.next_id += 1
        return res


class MyEnvironment(Environment):
    context_class = MyContext


def build_report_url(
    request: HttpRequest,
    *,
    report_info: dict[str, Any] | None = None,
    reported_url: str | None = None,
    kind: str | None = None,
    next_url: str | None = None,
    lookup_key: str | None = None,
):
    report_info = report_info or {}
    url_params = {
        'kind': kind if kind else report_info.get('kind', 'url'),
        'kind_data': reported_url if reported_url else report_info.get('url', request.build_absolute_uri(request.get_full_path())),
        'lookup_key': lookup_key if lookup_key else report_info.get('lookup_key', None),
        'next': next_url if next_url else request.get_full_path(),
    }
    if url_params['lookup_key'] is None:
        del url_params['lookup_key']
    return f'{reverse("plainui:report_content")}?{urlencode(url_params)}'


def environment(**options):
    env = MyEnvironment(**{'extensions': ['jinja2.ext.i18n', 'jinja2.ext.debug'], **options})
    env.globals.update(
        {
            'archive_mode': settings.ARCHIVE_MODE,
            'browser_type': browser_type,
            'build_report_url': build_report_url,
            'css_scope': css_scope,
            'active_theme': active_theme,
            'get_language': get_language,
            'get_messages': get_messages,
            'hub_absolute': hub_absolute,
            'hub_absolute_self': hub_absolute_self,
            'json_script': json_script,
            'num_of_notifications': num_of_notifications,
            'num_of_unread_messages': num_of_unread_messages,
            'num_of_pending_badges': num_of_pending_badges,
            'static': static,
            'url': url,
            'show_vars': show_vars,
            'unique_id': unique_id,
            'update_parameters': update_parameters,
            'translated_fields_for_field': translated_fields_for_field,
            'field_translation_languages': field_translation_languages,
            'now': timezone.now(),
            'calculate_luminance': calculate_luminance,
            'calculate_text_colour': calculate_text_colour,
            'lighten': lighten,
            'fetch_wiki_page': _fetch_wiki_page,
            'link_icon': link_icon,
        }
    )
    env.filters['strftdelta'] = custom_timedelta
    env.filters['strftdelta_short'] = custom_timedelta_short
    env.filters['strftimehm'] = custom_strftimehm
    env.filters['strftime'] = custom_strftime
    env.filters['strfdate'] = custom_strfdate
    env.filters['strfdates'] = custom_strfdates
    env.filters['weekday_abbrev'] = custom_weekday_abbrev
    env.install_gettext_callables(gettext, ngettext, newstyle=True)
    return env
