__all__ = (
    'PersonalMessageDeleteView',
    'PersonalMessageListView',
    'PersonalMessageSendView',
    'PersonalMessageShowView',
)

from django_ratelimit.decorators import ratelimit

from django.contrib import messages
from django.db.models import Q
from django.shortcuts import get_object_or_404, redirect
from django.urls import reverse
from django.utils.decorators import method_decorator
from django.utils.translation import gettext
from django.views.generic.base import TemplateView, View
from django.views.generic.edit import FormView

from core.markdown import render_markdown
from core.models import (
    DirectMessage,
)

from plainui.forms import (
    NewDirectMessageForm,
)
from plainui.views.utils import (
    ConferenceRequiredMixin,
)


class PersonalMessageListView(ConferenceRequiredMixin, TemplateView):
    require_user = True
    only_published = False
    template_name = 'plainui/personal_message_list.html.j2'
    sent = False

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['conf'] = self.conf

        PAGE_SIZE = 20
        dms = DirectMessage.objects.filter(conference=self.conf).order_by('-timestamp')
        context['sent_mode'] = self.sent
        try:
            start = int(self.request.GET['start'])
        except (ValueError, KeyError):
            start = 0
        if self.sent:
            dms = dms.filter(sender=self.request.user, deleted_by_sender=False).select_related('recipient', 'sender')
        else:
            dms = dms.filter(recipient=self.request.user, deleted_by_recipient=False).select_related('recipient', 'sender')

        context['start'] = start
        context['total'] = dms.count()
        context['msgs'] = dms[start : start + PAGE_SIZE]
        context['disable_share'] = True
        return context


@method_decorator(ratelimit(key='user', rate='3/m', method=ratelimit.UNSAFE), name='dispatch')
@method_decorator(ratelimit(key='ip', rate='5/m', method=ratelimit.UNSAFE), name='dispatch')
class PersonalMessageSendView(ConferenceRequiredMixin, FormView):
    require_user = True
    only_published = False
    template_name = 'plainui/personal_message_send.html.j2'
    form_class = NewDirectMessageForm

    def get_context_data(self, **kwargs):
        return {
            **super().get_context_data(**kwargs),
            'conf': self.conf,
            'disable_share': True,
        }

    def get_initial(self):
        initial = super().get_initial()
        initial['recipient'] = self.kwargs.get('recipient', '')
        initial['subject'] = self.request.POST.get('subject', self.request.GET.get('subject', ''))
        initial['in_reply_to'] = self.request.POST.get('in_reply_to', self.request.GET.get('in_reply_to', ''))
        return initial

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs['request'] = self.request
        kwargs['conf'] = self.conf
        kwargs['user'] = self.request.user
        return kwargs

    def get_success_url(self):
        return reverse('plainui:personal_message')

    def form_valid(self, form):
        in_reply_to = form.cleaned_data['in_reply_to']

        dm = DirectMessage(
            conference=self.conf,
            sender=self.request.user,
            recipient=form.cleaned_data['recipient'],
            in_reply_to_id=in_reply_to,
            subject=form.cleaned_data['subject'],
            body=form.cleaned_data['body'],
            deleted_by_recipient=self.request.user != form.cleaned_data['recipient'] and self.request.user.shadow_banned,
        )
        dm.save()
        if in_reply_to:
            DirectMessage.objects.filter(id=in_reply_to, recipient=self.request.user).update(has_responded=True)

        messages.success(self.request, gettext('Message sent.'))
        return super().form_valid(form)


class PersonalMessageShowView(ConferenceRequiredMixin, TemplateView):
    require_user = True
    only_published = False
    template_name = 'plainui/personal_message_show.html.j2'

    def get_context_data(self, msg_id, **kwargs):
        user = self.request.user
        context = super().get_context_data(**kwargs)
        context['conf'] = self.conf

        dms = DirectMessage.objects.filter(Q(sender=user, deleted_by_sender=False) | Q(recipient=user, deleted_by_recipient=False))
        context['msg'] = message = get_object_or_404(dms, pk=msg_id)
        if message.recipient == user and not message.was_read:
            DirectMessage.objects.filter(pk=message.pk).update(was_read=True)
        context['msg_body'] = render_markdown(self.conf, message.body)
        context['disable_share'] = True

        context['report_info'] = {'kind': 'pn', 'url': message.id}

        return context


class PersonalMessageDeleteView(ConferenceRequiredMixin, View):
    require_user = True
    only_published = False

    def get(self, request, **kwargs):
        return redirect(reverse('plainui:personal_message'))

    def post(self, request):
        DirectMessage.objects.filter(sender=request.user, pk=request.POST['id']).update(deleted_by_sender=True)
        DirectMessage.objects.filter(recipient=request.user, pk=request.POST['id']).update(deleted_by_recipient=True)
        DirectMessage.objects.filter(pk=request.POST['id'], deleted_by_sender=True, deleted_by_recipient=True).delete()

        messages.success(self.request, gettext('Message deleted.'))
        return redirect(reverse('plainui:personal_message'))
