__all__ = (
    'RedeemTokenAddToUserView',
    'RedeemTokenLoggedIn',
    'RedeemTokenUserCreateView',
    'RedeemTokenView',
)

from django_ratelimit.decorators import ratelimit

from django.contrib import messages
from django.contrib.auth import login
from django.contrib.auth import views as auth_views
from django.db import transaction
from django.http import HttpResponseRedirect
from django.shortcuts import redirect
from django.urls import reverse
from django.utils.decorators import method_decorator
from django.utils.translation import gettext
from django.views.generic.base import TemplateView, View
from django.views.generic.edit import FormView

from core.models import (
    ConferenceMember,
    ConferenceMemberTicket,
)
from core.models.ticket import TicketValidationError

from plainui.forms import (
    InputTokenForm,
    RedeemTokenAddToUserForm,
    RedeemTokenUserCreateForm,
)
from plainui.views.utils import (
    ConferenceRequiredMixin,
)


class RedeemTokenView(ConferenceRequiredMixin, TemplateView):
    template_name = 'plainui/redeem_token.html.j2'
    require_login = False
    require_conference_member = False

    def get(self, request, *args, **kwargs):
        if request.user.is_authenticated and self.conf.users.filter(user=self.request.user, has_ticket=True).exists():
            return HttpResponseRedirect(reverse('plainui:index'))

        return super().get(request, *args, **kwargs)

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(conf=self.conf, hide_header=True, **kwargs)
        token = self.request.GET.get('jwt')
        if token:
            form = InputTokenForm(conf=self.conf, user=self.request.user, data=self.request.GET)
            if not form.is_valid():
                context['step'] = 'token'
                context['form'] = form
                return context

            if self.request.user.is_authenticated:
                context['step'] = 'authenticated'
                context['form'] = form
                return context

            context['step'] = 'user'
            context['form_add'] = RedeemTokenAddToUserForm(conf=self.conf, initial={'token': form.cleaned_data['jwt']})
            context['form_create'] = RedeemTokenUserCreateForm(conf=self.conf, initial={'token': form.cleaned_data['jwt']})
        else:
            context['step'] = 'token'
            context['form'] = InputTokenForm(conf=self.conf, user=self.request.user)
        context['disable_share'] = True
        return context


@method_decorator(ratelimit(key='ip', rate='5/m', method=ratelimit.UNSAFE), name='dispatch')
@method_decorator(ratelimit(key='post:username', rate='5/m', method=ratelimit.UNSAFE), name='dispatch')
class RedeemTokenAddToUserView(ConferenceRequiredMixin, auth_views.LoginView):
    template_name = 'plainui/redeem_token.html.j2'
    require_login = False
    require_conference_member = False
    form_class = RedeemTokenAddToUserForm

    def get(self, request, *args, **kwargs):
        return redirect(reverse('plainui:redeem_token'))

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(**kwargs)
        context['conf'] = self.conf
        context['hide_header'] = True
        context['step'] = 'user'
        context['form_add'] = context['form']
        context['form_create'] = RedeemTokenUserCreateForm(conf=self.conf, initial={'token': context['form'].cleaned_data['token']})
        context['disable_share'] = True
        return context

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs['conf'] = self.conf
        return kwargs

    def form_valid(self, form: RedeemTokenAddToUserForm):
        try:
            with transaction.atomic():
                user = form.get_user()
                ConferenceMemberTicket.redeem_pretix_ticket(self.conf, user, form.cleaned_data['token'])
                ConferenceMember.objects.update_or_create(conference=self.conf, user=user, defaults={'has_ticket': True})
                login(self.request, user)
                return HttpResponseRedirect(self.get_success_url())
        except TicketValidationError as e:
            form.add_error(None, str(e))
            return self.render_to_response(self.get_context_data(form=form))

    def get_success_url(self):
        return reverse('plainui:index')


class RedeemTokenUserCreateView(ConferenceRequiredMixin, FormView):
    template_name = 'plainui/redeem_token.html.j2'
    require_login = False
    require_conference_member = False
    form_class = RedeemTokenUserCreateForm

    def get(self, request, *args, **kwargs):
        return redirect(reverse('plainui:redeem_token'))

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(**kwargs)
        context['conf'] = self.conf
        context['hide_header'] = True
        context['step'] = 'user'
        context['form_add'] = RedeemTokenAddToUserForm(conf=self.conf, initial={'token': context['form'].cleaned_data['token']})
        context['form_create'] = context['form']
        context['disable_share'] = True
        return context

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs['conf'] = self.conf
        return kwargs

    def form_valid(self, form: RedeemTokenUserCreateForm):
        try:
            with transaction.atomic():
                user = form.save()
                ConferenceMemberTicket.redeem_pretix_ticket(self.conf, user, form.cleaned_data['token'])
                ConferenceMember.objects.update_or_create(conference=self.conf, user=user, defaults={'has_ticket': True})
                login(self.request, user, backend='django.contrib.auth.backends.ModelBackend')
                return HttpResponseRedirect(self.get_success_url())
        except TicketValidationError as e:
            form.add_error(None, str(e))
            return self.render_to_response(self.get_context_data(form=form))

    def get_success_url(self):
        return reverse('plainui:index')


class RedeemTokenLoggedIn(ConferenceRequiredMixin, View):
    require_conference_member = False
    require_user = True

    def get(self, request, *args, **kwargs):
        return redirect(reverse('plainui:redeem_token'))

    def post(self, request, *args, **kwargs):
        token = request.POST['token']
        try:
            with transaction.atomic():
                if self.conf.end <= self.now:
                    raise TicketValidationError(gettext('The Conference is over!'))
                ConferenceMemberTicket.redeem_pretix_ticket(self.conf, request.user, token)
                ConferenceMember.objects.update_or_create(conference=self.conf, user=request.user, defaults={'has_ticket': True})
                return HttpResponseRedirect(reverse('plainui:index'))
        except TicketValidationError as e:
            messages.error(request, str(e))
            return HttpResponseRedirect(reverse('plainui:redeem_token'))
