__all__ = (
    'UserByUuidView',
    'UserView',
)

from django.db.models import Q
from django.shortcuts import get_object_or_404, redirect
from django.urls import reverse
from django.views.generic.base import TemplateView, View

from core.choices import UserBadgeVisibility
from core.models import (
    Event,
    EventParticipant,
    PlatformUser,
    UserBadge,
    UserContact,
)

from plainui.views.utils import (
    ConferenceRequiredMixin,
    session_get_favorite_events,
)


class UserView(ConferenceRequiredMixin, TemplateView):
    template_name = 'plainui/user.html.j2'

    def get_context_data(self, user_slug: str, **kwargs):
        user = self.request.user
        owned_badges = user.badges.all().values_list('badge__pk', flat=True) if user.is_authenticated else []

        context = super().get_context_data(**kwargs)
        context['conf'] = self.conf
        context['display_user'] = display_user = get_object_or_404(PlatformUser.objects.filter(slug=user_slug))
        conference_member = display_user.conferences.filter(conference=self.conf).first()
        context['description_html'] = None if not conference_member else conference_member.description_html
        context['owned_badges'] = owned_badges

        # Get public badges from user
        if user.is_authenticated:
            user_badges = UserBadge.objects.filter(user=user, accepted_by_user=True).values_list('badge_id', flat=True)
        else:
            user_badges = []

        contact = user.is_authenticated and UserContact.objects.filter(user=display_user, contact=user).first()
        if contact and not contact.pending:
            # Show Friend Badges
            context['badges'] = (
                UserBadge.objects.filter(user=display_user, accepted_by_user=True)
                .filter(
                    Q(visibility__in=[UserBadgeVisibility.PUBLIC, UserBadgeVisibility.FRIENDS, UserBadgeVisibility.CLUBFRIENDS])
                    | Q(badge__in=user_badges, visibility__in=[UserBadgeVisibility.CLUB, UserBadgeVisibility.CLUBFRIENDS])
                )
                .select_related('badge')
            )
        else:
            # Show only public badges
            context['badges'] = (
                UserBadge.objects.filter(user=display_user, accepted_by_user=True)
                .filter(
                    Q(visibility__in=[UserBadgeVisibility.PUBLIC])
                    | Q(badge__in=user_badges, visibility__in=[UserBadgeVisibility.CLUB, UserBadgeVisibility.CLUBFRIENDS])
                )
                .select_related('badge')
            )

        context['events'] = (
            Event.objects.conference_accessible(conference=self.conf)
            .filter(participants__participant=display_user)
            .filter(participants__role=EventParticipant.Role.SPEAKER)
            .filter(participants__is_public=True)
            .distinct()
            .order_by('schedule_start')
        )

        if len(context['events']) > 0:
            context['is_favorite_events'] = session_get_favorite_events(self.request.session, self.request.user)

        context['report_info'] = {'url': reverse('plainui:user', kwargs={'user_slug': display_user.slug})}

        return context


class UserByUuidView(ConferenceRequiredMixin, View):
    def get(self, *args, uuid, **kwargs):
        user = get_object_or_404(PlatformUser.objects.filter(uuid=uuid))
        return redirect(reverse('plainui:user', kwargs={'user_slug': user.slug}))
